from wake_t import (
    PlasmaStage,
    ActivePlasmaLens,
    Dipole,
    Quadrupole,
    Sextupole,
    GaussianPulse,
)

wake_t_conversion_rules = {
    "Dipole": Dipole,
    "Quadrupole": Quadrupole,
    "Sextupole": Sextupole,
    "Laser": GaussianPulse,
    "Plasma": PlasmaStage,
    "Plasma_Lens": ActivePlasmaLens,
}



from pydantic import BaseModel, ConfigDict
from ..units import UnitValue
import numpy as np


class FieldParameter(BaseModel):
    """
    FieldParameter class to represent a field parameter with a name and an optional value.

    Attributes
    ----------
    name (str):
        The name of the field parameter.
    value (UnitValue | None):
        The value of the field parameter, which can be a UnitValue or None.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    value: UnitValue | np.ndarray | list | None = None



import numpy as np

from ..units import UnitValue

from .FieldParameter import FieldParameter
from warnings import warn
from ..SDDSFile import SDDSFile, SDDS_Types


def write_SDDS_field_file(self, sddsindex: int = 0, ascii: bool = False) -> str:
    """
    Generate the field data in a format that is suitable for SDDS, based on the
    :class:`~nala.translatoru.utils.fields.field` object provided.
    This is then written to an SDDS file.
    The `field_type` parameter determines the format of the file.

    A warning is raised if the field type is not supported (perhaps elevate to a `NotImplementedError`?)

    Parameters
    ----------
    self: :class:`~nala.translator.utils.fields.field`
        The field object
    sddsindex: int
        Must be provided for :class:`~nala.translator.utils.SDDSFile.SddsFile` class
    ascii: bool, optional
        Convert to ascii?

    Returns
    -------
    str:
        The name of the SDDS field file.
    """
    sdds_filename = self._output_filename(extension=".sdds")
    sddsfile = SDDSFile(index=sddsindex, ascii=ascii)
    zdata = self.z_values
    tdata = self.t_values
    if self.field_type == "LongitudinalWake":
        wzdata = self.Wz.value.val
        cnames = ["z", "t", "Wz"]
        cunits = ["m", "s", "V/C"]
        ccolumns = [
            zdata,
            tdata,
            wzdata,
        ]
    elif self.field_type == "TransverseWake":
        wxdata = self.Wx.value.val
        wydata = self.Wy.value.val
        ccolumns = np.array(
            [
                zdata,
                tdata,
                wxdata,
                wydata,
            ]
        )
        cnames = ["z", "t", "Wx", "Wy"]
        cunits = ["m", "s", "V/C/m", "V/C/m"]
    elif self.field_type == "3DWake":
        wxdata = self.Wx.value.val
        wydata = self.Wy.value.val
        wzdata = self.Wz.value.val
        ccolumns = np.array(
            [
                zdata,
                tdata,
                wxdata,
                wydata,
                wzdata,
            ]
        )
        cnames = ["z", "t", "Wx", "Wy", "Wz"]
        cunits = ["m", "s", "V/C/m", "V/C/m", "V/C"]
    elif self.field_type == "1DElectroDynamic":
        ezdata = self.Ez.value.val
        cnames = ["z", "Ez"]
        cunits = ["m", "V"]
        ccolumns = [
            zdata,
            ezdata,
        ]
    else:
        warn(f"Field type {self.field_type} not supported for SDDS")
        return
    if ccolumns is not None:
        ctypes = [SDDS_Types.SDDS_DOUBLE for _ in ccolumns]
        csymbols = ["" for _ in ccolumns]
        sddsfile.add_columns(cnames, ccolumns, ctypes, cunits, csymbols)
        sddsfile.write_file(sdds_filename)
    return sdds_filename


def read_SDDS_field_file(self, filename: str, field_type: str):
    """
    Read an SDDS field file and convert it into a :class:`nala.translator.utils.fields.field` object.
    Only works for wakefield files.

    Parameters
    ----------
    self: :class:`~nala.translator.utils.fields.field`
        The field object to be updated.
    filename: str
        The path to the SDDS field file
    field_type: str
        The name of the field, see :attr:`~nala.translator.utils.fields.allowed_fields`

    Returns
    -------
    None

    Raises
    ------
    NotImplementedError:
        if a given `field_type` is not implemented
    """
    self.reset_dicts()
    setattr(self, "field_type", field_type)
    try:
        elegantObject = SDDSFile(index=1, ascii=True)
    except Exception:
        elegantObject = SDDSFile(index=1, ascii=False)
    elegantObject.read_file(filename, page=-1)
    if field_type in ["LongitudinalWake", "3DWake", "TransverseWake"]:
        for key, val in elegantObject._columns.items():
            data = np.array(val.data)
            if val.unit == "m":
                setattr(
                    self,
                    "z",
                    FieldParameter(name="z", value=UnitValue(data, units=val.unit)),
                )
            elif val.unit == "s":
                setattr(
                    self,
                    "t",
                    FieldParameter(
                        name="t", value=UnitValue(data, units="m")
                    ),
                )
            elif val.unit == "V/C":
                setattr(
                    self,
                    "Wz",
                    FieldParameter(name="Wz", value=UnitValue(data, units="V/C")),
                )
            elif val.unit == "V/C/m":
                setattr(
                    self,
                    "Wx",
                    FieldParameter(name="Wx", value=UnitValue(data, units="V/C/m")),
                )
                setattr(
                    self,
                    "Wy",
                    FieldParameter(name="Wy", value=UnitValue(data, units="V/C/m")),
                )
            else:
                raise ValueError(f"Unit {val.unit} not recognised in {filename}")
    else:
        raise NotImplementedError(
            f"{field_type} loading not implemented for SDDS files"
        )



import numpy as np
from pydantic import BaseModel
from typing import Dict
from ...utils.elegant import SDDSFile
import nala.models.element as NALA_elements
from nala.models.elementList import SectionLattice, MachineLayout, ElementList
from ...utils.elegant.sdds_classes_APS import SDDS_Floor, SDDS_Params

class ElegantLatticeImporter(BaseModel):

    params_file: str
    """Name of ELEGANT parameters file"""

    floor_file: str
    """Name of ELEGANT floor file"""

    elegant_data: Dict = {}
    """Dictionary containing data about the ELEGANT lattice"""

    floor_data: Dict = {}
    """Dictionary containing floor positions for the ELEGANT lattice"""

    elements: Dict = {}
    """Dictionary containing converted 
    :class:`~nala.models.element.Element` objects"""

    def create_element_dictionary(self):
        params = SDDS_Params(self.params_file)
        self.elegant_data, filenames = params.create_element_dictionary()
        return self.elegant_data, filenames

    def update_floor_coordinates(self):
        flr = SDDS_Floor()
        flr.import_sdds_floor_file(self.floor_file)
        self.floor_data = flr.data

        i = 0
        for k, v in self.floor_data.items():
            if i == 0:
                pass
            else:
                prevind = list(self.floor_data.keys()).index(k) - 1
                self.floor_data[k].update(
                    {
                        "start": list(self.floor_data.values())[prevind]["end"],
                        "start_rotation": list(self.floor_data.values())[prevind]["end_rotation"]
                    }
                )
            i += 1

    def create_nala_element_dictionary(self):
        if not self.elegant_data:
            self.create_element_dictionary()
        if not self.floor_data:
            self.update_floor_coordinates()
        self.elements = {}

        def calculate_middle_from_start(start_pos, length, physical_angle, rotation):
            """Calculate middle position from start position, accounting for rotation and bending."""
            start = np.array(start_pos)

            # Calculate local offset from start to middle
            if abs(physical_angle) > 1e-9:
                # Bent element - arc geometry
                mx_local = length * (1 - np.cos(physical_angle)) / (2 * physical_angle)
                my_local = 0
                mz_local = length * np.sin(physical_angle) / (2 * physical_angle)
            else:
                # Straight element
                mx_local = 0
                my_local = 0
                mz_local = length / 2.0

            # Apply only the yaw rotation (horizontal bending)
            if len(rotation) >= 3 and abs(rotation[2]) > 1e-12:
                yaw = rotation[2]
                # Rotation around Y axis
                cos_yaw = np.cos(yaw)
                sin_yaw = np.sin(yaw)

                mx_rotated = mx_local * cos_yaw + mz_local * sin_yaw
                my_rotated = my_local  # No change in y for yaw rotation
                mz_rotated = -mx_local * sin_yaw + mz_local * cos_yaw

                local_offset = np.array([mx_rotated, my_rotated, mz_rotated])
            else:
                local_offset = np.array([mx_local, my_local, mz_local])

            # Calculate middle position
            middle = start + local_offset
            return middle

        for k, v in self.elegant_data.items():
            if k in self.floor_data:
                vtype = v["hardware_type"]
                if "drift" not in vtype.lower():
                    if "l" in v and v["l"] > 0:
                        # Get physical angle for bent elements
                        physical_angle = 0.0
                        if v["hardware_type"].lower() == "dipole" and "angle" in v:
                            physical_angle = v["angle"]

                        # Calculate middle position properly
                        centre = calculate_middle_from_start(
                            start_pos=self.floor_data[k]["start"],
                            length=v["l"],
                            physical_angle=physical_angle,
                            rotation=self.floor_data[k]["start_rotation"]
                        )
                    else:
                        # Zero length element - middle is same as start
                        centre = np.array(self.floor_data[k]["start"])
                        physical_angle = 0.0

                    v = self._convert_k_to_kl(v)
                    v = self._convert_ele_phase_to_phase(v)

                    rotation = self.floor_data[k]["start_rotation"]
                    if "physical" in v:
                        v["physical"].update(
                            {
                                "middle": centre.tolist(),
                                "global_rotation": rotation,
                            }
                        )
                    else:
                        v["physical"] = {
                            "middle": centre.tolist(),
                            "global_rotation": rotation,
                        }

                    # Add physical_angle for bent elements
                    if abs(physical_angle) > 1e-9:
                        v["physical"]["physical_angle"] = physical_angle

                    self.elements.update({k: getattr(NALA_elements, vtype)(**v)})

    def create_section(self, section: Dict) -> Dict[str, SectionLattice]:
        if not self.elements:
            self.create_nala_element_dictionary()
        secname = list(section.keys())
        assert len(secname) == 1
        secelements = list(section.values())[0]
        assert len(secelements) >= 2
        appending = False
        order = []
        elems = {}
        for name, elem in self.elements.items():
            if name == secelements[0]:
                appending = True
            elif name == secelements[1]:
                appending = False
            if appending:
                order.append(name)
                elems.update({name: elem})
        if not order:
            raise KeyError(f"element {secelements[0]} not found in lattice; could not construct section")
        seclat = SectionLattice(order=order, elements=ElementList(elements=elems), name=secname[0])
        return {secname[0]: seclat}

    def create_layout(self, name: str, sections: Dict) -> MachineLayout:
        layout_sections = {}
        for secname, secpos in sections.items():
            layout_sections.update(self.create_section({secname: secpos}))
        return MachineLayout(
            name=name,
            sections={k: v for k, v in zip(list(layout_sections.keys()), list(layout_sections.values()))}
        )

    @staticmethod
    def _convert_k_to_kl(v) -> dict:
        multi = {}
        if "angle" in v:
            v["k0"] = v["angle"] / float(v["magnetic"]["length"])
            v["physical"]["physical_angle"] = -v["angle"]
        for n in range(0, 9):
            if f"k{n}" in v and ("length" in v["magnetic"] or "length" in v["physical"]):
                try:
                    knl = float(v[f"k{n}"]) * float(v["magnetic"]["length"])
                except KeyError:
                    knl = float(v[f"k{n}"]) * float(v["physical"]["length"])
                multi.update({f"K{n}L": {"order": n, "normal": knl}})
                del v[f"k{n}"]
        if "magnetic" in v:
            v["magnetic"].update({"multipoles": multi})
        return v

    @staticmethod
    def _convert_ele_phase_to_phase(v) -> dict:
        if "cavity" in v:
            if "phase" in v["cavity"]:
                v["cavity"]["phase"] = (90 - v["cavity"]["phase"])
        return v

    @staticmethod
    def import_sdds_params_file(filename: str, page: int = 0) -> list:
        elegantObject = SDDSFile(index=1)
        elegantObject.read_file(filename, page=page)
        elegantData = elegantObject.data
        return elegantData


from typing import Any, List
from .base import BaseElementTranslator
from nala.models.laser import LaserElement


class LaserTranslator(BaseElementTranslator):
    """
    Translator class for converting a :class:`~nala.models.element.Laser` element instance into a string or
    object that can be understood by various simulation codes.
    """

    laser: LaserElement
    """Laser element class."""

    supported_pulses: List = [
        'gaussian',
        'laguerre-gaussian',
        'flattened-gaussian',
        # 'file'
    ]
    """Types of laser pulses that can be supported."""

    additional_attrs: List = [
        "focal_position",
        "wavelength",
        "cep_phase",
        "polarization",
    ]
    """Additional laser attributes."""

    def to_wake_t(self) -> Any:
        """
        Create a Wake-T laser element object based on the attributes of this element.

        Returns
        -------
        wake_t.LaserPulse
            Wake-T laser element object

        Raises
        ------
        ValueError
            If the laser model is not supported; note that not all models are implemented yet
        """
        from wake_t.physics_models.laser.laser_pulse import (
            GaussianPulse,
            LaguerreGaussPulse,
            FlattenedGaussianPulse,
            # SummedPulse,
            # OpenPMDPulse,
        )
        additional_dict = {
            self._convertKeyword_WakeT(param): getattr(self.laser, param) for param in self.additional_attrs
        }
        if self.profile_type == "gaussian":
            obj = GaussianPulse(
                self.laser.initial_position,
                self.laser.amplitude,
                self.laser.waist,
                self.laser.pulse_duration_fwhm,
                **additional_dict,
            )
        elif self.profile_type == "laguerre-gaussian":
            obj = LaguerreGaussPulse(
                self.laser.initial_position,
                self.laser.laguerre_polynomial_order_p,
                self.laser.amplitude,
                self.laser.waist,
                self.laser.pulse_duration_fwhm,
                **additional_dict,
            )
        elif self.profile_type == "flattened-gaussian":
            obj = FlattenedGaussianPulse(
                self.laser.initial_position,
                self.laser.amplitude,
                self.laser.waist,
                self.laser.pulse_duration_fwhm,
                N=self.laser.flatness,
                **additional_dict,
            )
        else:
            raise ValueError(
                f"Invalid laser profile type {self.laser.profile_type}. "
                f"Supported models are {self.supported_pulses}."
            )
        return obj


import numpy as np
from .sdds import write_SDDS_field_file
from warnings import warn
from collections import Counter
from .FieldParameter import FieldParameter
from ..units import UnitValue
import re

d = ",!?/&-:;@'\n \t"


def write_opal_field_file(
    self,
    frequency: float = None,
    radius: float = None,
    fourier: int = 100,
    orientation: str = None,
):
    """
    Generate the field data in a format that is suitable for OPAL, based on the
    :class:`~nala.translator.utils.fields.field` object provided.

    See the `OPAL manual`_ for more details.

    This is then written to a text file.
    The `field_type` parameter determines the format of the file.

    A warning is raised if the field type is not supported (perhaps elevate to a `NotImplementedError`?

    .. _OPAL manual: https://amas.web.psi.ch/opal/Documentation/master/OPAL_Manual.html

    Parameters
    ----------
    self: :class:`~nala.translator.utils.fields.field`
        The field object
    frequency: float | None
        The frequency of the field
    radius: float | None
        The radius of the field, for 1D axially symmetric fields; defaults to 0.1
    fourier: int = 100
        Number of fourier coefficients
    orientation: str | None
        Orientation of the field (for 2D)

    Returns
    -------
    str | None:
        The name of the field file.
        Will return None if required parameters for certain fields are not provided.

    Raises
    ------
    Warning:
        if too many or not enough fourier components are provided; will default to 1/100
    Warning:
        if the magnet radius is not provided; will default to 0.1
    Warning:
        if the RF frequency, radius or orientation is not provided for
        `1DElectroDynamic` or `2DElectrodynamic`, return None
    Warning:
        if trying to use wakefields; these are tricky to get working with OPAL
    Warning:
        if a given `field_type` is not supported.
    """
    length = self.length
    opal_file = self._output_filename(extension=".opal")
    data = None
    header = None
    fourier_ratio = fourier / length
    if self.field_type != "2DElectroDynamic":
        if fourier_ratio > 1.0:
            warn("Too many fourier components provided; defaulting to 1/100")
            fourier = int(length / 100)
        elif fourier_ratio < 0.01:
            warn("Not enough fourier components provided; defaulting to 1/100")
            fourier = int(length / 100)
    if self.field_type == "1DMagnetoStatic":
        zmin = min(self.z.value.val) * 1e2
        zmax = max(self.z.value.val) * 1e2
        data = self.Bz.value.val
        head = ["1DMagnetoStatic", str(fourier)]
        if radius is None:
            warn("Magnet radius not provided; defaulting to 10cm")
            radius = 0.1
        zvals = [str(zmin), str(zmax), str(length-1)]
        rvals = [str(0), str(radius * 100), fourier]
        header = [head, zvals, rvals]
    elif self.field_type == "1DElectroDynamic":
        if frequency is None:
            warn("RF Frequency not provided to field class")
            return
        head = ["AstraDynamic", str(fourier)]
        freq = [str(frequency * 1e-6)]
        zdata = self.z.value.val
        ezdata = self.Ez.value.val
        data = np.transpose([zdata, ezdata])
        header = [head, freq]
    elif self.field_type == "2DElectroDynamic":
        if frequency is None:
            warn("RF Frequency not provided to field class")
            return
        if radius is None:
            warn("Radius not provided to field class")
            return
        if orientation is None:
            warn("Orientation not provided to field class")
            return
        orient = self.orientation
        ezvals = self.Ez.value.val
        ervals = self.Er.value.val
        eabsvals = self.Ex.value.val
        brvals = self.Br.value.val
        shape = len(ezvals)
        count = Counter(ezvals)
        repeating_floats = {key: value for key, value in count.items() if value > 1}
        zlen = int(len(repeating_floats) - 1)
        rlen = int((shape / len(repeating_floats)) - 1)
        head = ["2DDynamic", orient]
        leng = ["0.0", str(self.length * 100), str(zlen)]
        freq = [str(frequency * 1e-6)]
        rad = ["0.0", str(self.radius * 100), str(rlen)]
        data = np.transpose([ezvals, ervals, eabsvals, brvals])
        header = [head, leng, freq, rad]
    elif "wake" in self.field_type.lower():
        warn(f"Field type {self.field_type} defaulting to SDDS type; use with caution")
        return write_SDDS_field_file(self)
    else:
        warn(f"Field type {self.field_type} not supported for OPAL")
    if data is not None:
        with open(f"{opal_file}", "w") as f:
            for h in header:
                f.write(" ".join([str(x) for x in h]) + "\n")
            for dat in data:
                if not type(dat) in [list, np.ndarray]:
                    dat = [dat]
                f.write(" ".join([str(x) for x in dat]) + "\n")
    return opal_file


def read_opal_field_file(
    self,
    filename: str,
    field_type: str,
    cavity_type: str | None = None,
    frequency: float | None = None,
):
    """
    Read an OPAL field file and convert it into a :class:`nala.translator.utils.fields.field` object

    Parameters
    ----------
    self: :class:`~nala.translator.utils.fields.field`
        The field object to be updated.
    filename: str
        The path to the OPAL field file
    field_type: str
        The name of the field, see :attr:`~nala.translator.utils.fields.allowed_fields`
    cavity_type: str, optional
        The type of RF cavity, see :attr:`~nala.translator.utils.fields.allowed_cavities`
    frequency: float, optional
        The frequency of the RF cavity.

    Returns
    -------
    None

    Raises
    ------
    ValueError:
        if the cavity `field_type` contains the string `Electro` and `cavity_type` is not provided
    ValueError:
        if the cavity `field_type` contains the string `Electro` and `frequency` is not provided
    NotImplementedError:
        if a given `field_type` is not implemented
    """
    self.reset_dicts()
    setattr(self, "field_type", field_type)
    if "Electro" in field_type:
        if cavity_type is None:
            raise ValueError(f"cavity_type must be provided for {field_type}")
        else:
            setattr(self, "cavity_type", cavity_type)
        if frequency is None:
            raise ValueError(f"frequency must be provided for {field_type}")
        else:
            setattr(self, "frequency", frequency)
    if field_type == "2DElectroDynamic":
        with open(filename) as f:
            rl = f.readlines()
            setattr(self, "length", float(rl[1].split(" ")[1]) * 1e-2)
            setattr(self, "frequency", float(rl[2]) * 1e6)
            setattr(self, "radius", float(rl[3].split(" ")[1]) * 1e-2)
            setattr(self, "orientation", rl[0].split(" ")[1].strip("\n"))
        fdat = np.loadtxt(filename, skiprows=4)
        zfull = np.tile(np.linspace(0, self.length, int(rl[1].split(" ")[2]) + 1), int(rl[3].split(" ")[2]) + 1)
        values = np.arange(0, self.radius, self.radius / (int(rl[3].split(" ")[2]) + 1))
        rpattern = np.repeat(values, int(rl[1].split(" ")[2]) + 1)
        setattr(
            self,
            "z",
            FieldParameter(name="z", value=UnitValue(zfull, units="m")),
        )
        setattr(
            self,
            "r",
            FieldParameter(name="r", value=UnitValue(rpattern, units="m")),
        )
        setattr(
            self,
            "Ez",
            FieldParameter(name="Ez", value=UnitValue(fdat[::, 0], units="V/m")),
        )
        setattr(
            self,
            "Er",
            FieldParameter(name="Er", value=UnitValue(fdat[::, 1], units="V/m")),
        )
        setattr(
            self,
            "Ex",
            FieldParameter(name="Ex", value=UnitValue(fdat[::, 2], units="V/m")),
        )
        setattr(
            self,
            "Br",
            FieldParameter(name="Br", value=UnitValue(fdat[::, 3], units="T")),
        )
    elif field_type == "1DMagnetoStatic":
        with open(filename) as f:
            rl = f.readlines()
            if rl[0].split(" ")[0] not in ["1DMagnetoStatic"]:
                raise NotImplementedError(
                    f"{rl[0].split(' ')[0]} field type not implemented for OPAL fields"
                )
            setattr(self, "fourier", int(rl[0].split(" ")[1]))
            setattr(self, "radius", float(rl[2].split(" ")[1]) * 1e-2)
            zstart = float(rl[1].split(" ")[0]) * 1e-2
            zend = float(rl[1].split(" ")[1]) * 1e-2
        fdat = np.loadtxt(filename, skiprows=3)
        zvals = np.linspace(zstart, zend, len(fdat))
        setattr(self, "z", FieldParameter(name="z", value=UnitValue(zvals, units="m")))
        setattr(
            self,
            "Bz",
            FieldParameter(name="Bz", value=UnitValue(fdat / max(fdat), units="T")),
        )
    elif field_type == "1DElectroDynamic":
        with open(filename) as f:
            rl = f.readlines()
            if rl[0].split(" ")[0] not in ["AstraDynamic"]:
                raise NotImplementedError(
                    f"{rl[0].split(' ')[0]} field type not implemented for OPAL fields"
                )
            setattr(self, "fourier", int(rl[0].split(" ")[1]))
            setattr(self, "frequency", float(rl[1]) * 1e6)
        fdat = np.loadtxt(filename, skiprows=2)
        setattr(
            self, "z", FieldParameter(name="z", value=UnitValue(fdat[::, 0], units="m"))
        )
        setattr(
            self,
            "Ez",
            FieldParameter(
                name="Ez", value=UnitValue(fdat[::, 1] / max(fdat[::, 1]), units="V/m")
            ),
        )



# import unittest
# from nala import NALA
# from PAdantic.models.element import Quadrupole
# from PAdantic.models.magnetic import Quadrupole_Magnet
# from PAdantic.models.physical import PhysicalElement, Position
#
#
# class TestPAdantic(unittest.TestCase):
#     def setUp(self):
#         self.q1 = Quadrupole(
#             name="QUAD1",
#             machine_area="FODO",
#             magnetic=Quadrupole_Magnet(
#                 length=0.1,
#                 k1l=1.0,
#             ),
#             physical=PhysicalElement(
#                 middle=Position(
#                     x=0,
#                     y=0,
#                     z=0.1,
#                 ),
#                 length=0.1,
#             ),
#         )
#         self.q2 = Quadrupole(
#             name="QUAD2",
#             machine_area="NODO",
#             magnetic=Quadrupole_Magnet(
#                 length=0.1,
#                 k1l=1.0,
#             ),
#             physical=PhysicalElement(
#                 middle=Position(
#                     x=0,
#                     y=0,
#                     z=0.5,
#                 ),
#                 length=0.1,
#             ),
#         )
#         self.sections = {
#             "sections": {
#                 "FODO": ["QUAD1"],
#                 "NODO": ["QUAD2"],
#             }
#         }
#         self.layouts = {
#             "layouts": {
#                 "line1": ["FODO", "NODO"],
#             }
#         }
#
#         self.machine = PAdantic(
#             layout=self.layouts,
#             section=self.sections,
#             element_list=[self.q1, self.q2],
#         )
#
#     def test_get_elements_s_pos(self):
#         elements_s_pos = self.machine.get_elements_s_pos()
#         self.assertIsInstance(elements_s_pos, dict)
#         self.assertGreater(len(elements_s_pos), 0)
#
#     def test_get_dipoles(self):
#         dipoles = self.machine.get_dipoles()
#         self.assertIsInstance(dipoles, list)
#         self.assertEqual(len(dipoles), 0)
#
#     def test_get_quadrupoles(self):
#         quadrupoles = self.machine.get_quadrupoles()
#         self.assertIsInstance(quadrupoles, list)
#         self.assertGreater(len(quadrupoles), 0)
#
#     def test_get_element_with_two_paths_fails_when_path_is_not_provided(self):
#         layouts = {
#             "layouts": {
#                 "line1": ["FODO", "NODO"],
#                 "line2": ["NODO", "FODO"],
#             }
#         }
#         machine = PAdantic(
#             layout=layouts,
#             section=self.sections,
#             element_list=[self.q1, self.q2],
#         )
#         with self.assertRaises(Exception) as context:
#             _ = machine.get_quadrupoles()
#             self.assertEqual(context.msg, '"default_layout" = None is not defined')
#
#     def test_get_element_with_two_paths_and_path_defined(self):
#         layouts = {
#             "layouts": {
#                 "line1": ["FODO", "NODO"],
#                 "line2": ["NODO", "XODO"],
#             }
#         }
#         q3 = Quadrupole(
#                     name="QUAD3",
#                     machine_area="XODO",
#                     magnetic=Quadrupole_Magnet(
#                         length=0.1,
#                         k1l=1.0,
#                     ),
#                     physical=PhysicalElement(
#                         middle=Position(
#                             x=0,
#                             y=0,
#                             z=0.6,
#                         ),
#                         length=0.1,
#                     ),
#                 )
#         sections = {
#                     "sections": {
#                         "FODO": ["QUAD1"],
#                         "NODO": ["QUAD2"],
#                         "XODO": ["QUAD3"],
#                     }
#                 }
#         machine = PAdantic(
#             layout=layouts,
#             section=sections,
#             element_list=[self.q1, self.q2, q3],
#         )
#         quadrupoles = machine.get_quadrupoles(path="line1")
#         self.assertIsInstance(quadrupoles, list)
#         self.assertGreater(len(quadrupoles), 0)
#         quadrupoles = machine.get_quadrupoles(path="line2")
#         self.assertIsInstance(quadrupoles, list)
#         self.assertGreater(len(quadrupoles), 0)



# python
import pytest
import numpy as np
from nala.models.physical import (
    Position,
    Rotation,
    ElementError,
    PhysicalElement,
)


def test_position_operations():
    p1 = Position(x=1, y=2, z=3)
    p2 = Position(x=4, y=5, z=6)
    assert (p1 + p2) == Position(x=5, y=7, z=9)
    assert (p2 - p1) == Position(x=3, y=3, z=3)
    assert p1.dot(p2) == 32
    assert p1.length() == pytest.approx(3.7417, rel=1e-3)

def test_rotation_operations():
    r1 = Rotation(phi=0.1, psi=0.2, theta=0.3)
    r2 = Rotation(phi=0.4, psi=0.5, theta=0.6)
    assert all(
        [np.isclose(x, y) for x, y in zip((r1 + r2).model_dump(), Rotation(phi=0.5, psi=0.7, theta=0.9).model_dump())])
    assert all(
        [np.isclose(x, y) for x, y in zip((r2 - r1).model_dump(), Rotation(phi=0.3, psi=0.3, theta=0.3).model_dump())])
    assert abs(r1) == Rotation(phi=0.1, psi=0.2, theta=0.3)

def test_element_error_initialization():
    error = ElementError(position=[1, 2, 3], rotation=[0.1, 0.2, 0.3])
    assert error.position == Position(x=1, y=2, z=3)
    assert error.rotation == Rotation(theta=0.1, phi=0.2, psi=0.3)

def test_physical_element_properties():
    pe = PhysicalElement(middle=[1, 2, 3], length=10, global_rotation=[0.1, 0.2, 0.3])
    assert pe.start.x < pe.middle.x
    assert pe.end.z > pe.middle.z
    assert pe.rotation_matrix is not None


from pydantic import Field

from .baseModels import IgnoreExtra


class ElectricalElement(IgnoreExtra):
    """
    Electrical info model.
    """

    minI: float = Field(alias="min_i", default=0)
    """Minimum current that can be set [A]."""

    maxI: float = Field(alias="max_i", default=0)
    """Maximum current that can be set [A]."""

    read_tolerance: float = Field(alias="ri_tolerance", default=0.1)
    """Tolerance on read current [A]."""



from pydantic import BaseModel, ConfigDict, computed_field
from typing import List
import numpy as np
from ...utils.classes import getGrids
from ...utils.functions import chop


class gpt_ccs(BaseModel):

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
        validate_assignment=True,
        populate_by_name=True,
    )

    name: str

    position: List[float] = [0.0, 0.0, 0.0]

    rotation: List[float] = [0.0, 0.0, 0.0]

    intersect: float = 0.0

    @computed_field
    @property
    def psi(self) -> float:
        return self.rotation[0]

    @computed_field
    @property
    def phi(self) -> float:
        return self.rotation[1]

    @computed_field
    @property
    def theta(self) -> float:
        return self.rotation[2]

    @computed_field
    @property
    def x(self) -> float:
        return self.position[0]

    @computed_field
    @property
    def y(self) -> float:
        return self.position[1]

    @computed_field
    @property
    def z(self) -> float:
        return self.position[2]

    def relative_position(
        self, position: np.ndarray | list, rotation: np.ndarray | list
    ) -> tuple:
        x, y, z = position
        pitch, yaw, roll = rotation
        length = np.sqrt((x - self.x) ** 2 + (y - self.y) ** 2 + (z - self.z) ** 2)
        finalrot = np.array([pitch - self.psi, yaw - self.phi, roll - self.theta])
        finalpos = np.array(
            [0, 0, abs(self.intersect) + length]
        )
        return finalpos, finalrot

    @property
    def name_as_str(self):
        return '"' + self.name + '"'

    def ccs_text(self, position, rotation):
        finalpos, finalrot = self.relative_position(position, rotation)
        x, y, z = finalpos
        psi, phi, theta = finalrot
        ccs_label = ""
        value_text = ""
        if abs(x) > 0:
            ccs_label += "x"
            value_text += "," + str(x)
        if abs(y) > 0:
            ccs_label += "y"
            value_text += "," + str(y)
        if abs(z) > 0:
            ccs_label += "z"
            value_text += "," + str(z)
        if abs(psi) > 0:
            ccs_label += "X"
            value_text += "," + str(psi)
        if abs(phi) > 0:
            ccs_label += "Y"
            value_text += "," + str(phi)
        if abs(theta) > 0:
            ccs_label += "Z"
            value_text += "," + str(theta)
        if ccs_label == "" and value_text == "":
            ccs_label = "z"
            value_text = "," + str(0)
        return '"' + ccs_label + '"', value_text.strip(",")

    def gpt_coordinates(self, position: list | np.ndarray, rotation: list | np.ndarray) -> str:
        """
        Get the GPT coordinates for a given position and rotation

        Parameters
        ----------
        position: list | np.ndarray
            The lattice position.
        rotation: float
            The element rotation

        Returns
        -------
        str
            A GPT-formatted position string.
        """
        x, y, z = chop(position, 1e-6)
        psi, phi, theta = rotation
        output = ""
        for c in [-x, y, z]:
            output += str(c) + ", "
        output += "cos(" + str(theta) + "), 0, -sin(" + str(theta) + "), 0, 1 ,0"
        return output

class gpt_element(BaseModel):
    """
    Generic class for generating headers for GPT.
    """

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
        validate_assignment=True,
        populate_by_name=True,
    )

    objectname: str = ""

    objecttype: str = ""

    exclude: List = ["exclude", "objectname", "objecttype", "particle_definition"]

    def write_GPT(self, *args, **kwargs) -> str:
        """
        Write the text for the GPT namelist based on its
        :attr:`~objectdefaults`, :attr:`~objectname`.

        Returns
        -------
        str
            GPT-compatible string representing the namelist
        """
        output = str(self.objectname) + "("
        for key, val in self.model_dump().items():
            if key not in self.exclude and val is not None:
                k = key.lower()
                output += str(getattr(self, k)) + ", "
        output = output[:-2]
        output += ");\n"
        return output


class gpt_setfile(gpt_element):
    """
    Class for setting filenames in GPT via `setfile`.
    """

    objectname: str = "setfile"
    """Name of object"""

    objecttype: str = "gpt_setfile"
    """Type of object"""

    set: str = '"beam"'
    """Name of beam"""

    filename: str = '"output.gdf"'
    """Output filename"""

    particle_definition: str = '"gdf"'
    """Name of particle definition"""

    time: float = None
    """Mean time coordinate of particle #TODO does this belong here?"""

    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.exclude.extend(["time"])


class gpt_charge(gpt_element):
    """
    Class for generating the `settotalcharge` namelist for GPT.
    """

    set: str = '"beam"'
    """Name of beam for `settotalcharge`"""

    charge: float = 0.0
    """Bunch charge"""

    objectname: str = "settotalcharge"
    """Name of object"""

    objecttype: str = "gpt_charge"
    """Type of object"""


    def write_GPT(self, *args, **kwargs) -> str:
        output = str(self.objectname) + "("
        output += str(self.set) + ","
        output += str(-1 * abs(self.charge)) + ");\n"
        return output


class gpt_setreduce(gpt_element):
    """
    Class for reducing the number of particles via `setreduce`.

    """

    set: str = '"beam"'
    """Name of the beam for `setreduce`"""

    setreduce: int = 1
    """Factor by which to reduce the number of particles"""

    objectname: str = "setreduce"
    """Name of object"""

    objecttype: str = "gpt_setreduce"
    """Type of object"""


    def write_GPT(self, *args, **kwargs) -> str:
        output = str(self.objectname) + "("
        output += str(self.set) + ","
        output += str(self.setreduce) + ");\n"
        return output


class gpt_accuracy(gpt_element):
    """
    Class for setting the accuracy of tracking via `accuracy` in GPT.
    """

    objectname: str = "accuracy"
    """Name of object"""

    objecttype: str = "gpt_accuracy"
    """Type of object"""

    accuracy: int = 6
    """Accuracy for GPT tracking"""

    def write_GPT(self, *args, **kwargs) -> str:
        output = (
            "accuracy(" + str(self.accuracy) + ");\n"
        )  # 'setrmacrodist(\"beam\","u",1e-9,0) ;\n'
        return output


class gpt_spacecharge(gpt_element):
    """
    Class for preparing space charge calculations in GPT via `spacecharge`.
    """

    grids: getGrids = None
    """Class for calculating the required number of space charge grids"""

    ngrids: int | None = None
    """Number of space charge grids"""

    space_charge_mode: str | None = None
    """Space charge mode ['2D', '3D']"""

    cathode: bool = False
    """Flag indicating whether the bunch was emitted from a cathode"""

    objectname: str = "spacecharge"
    """Name of object"""

    objecttype: str = "gpt_spacecharge"
    """Type of object"""

    def model_post_init(self, __context):
        super().model_post_init(__context)
        self.grids = getGrids()
        self.exclude.extend(["cathode", "grids", "ngrids", "space_charge_mode"])

    def write_GPT(self, *args, **kwargs) -> str:
        output = ""
        if isinstance(self.space_charge_mode, str) and self.cathode:
            if self.ngrids is None:
                self.ngrids = self.grids.getGridSizes(
                    (self.npart / self.sample_interval)
                )
            output += 'spacecharge3Dmesh("Cathode","RestMaxGamma",1000);\n'
        elif (
            isinstance(self.space_charge_mode, str)
            and self.space_charge_mode.lower() == "3d"
        ):
            output += "Spacecharge3Dmesh();\n"
        elif (
            isinstance(self.space_charge_mode, str)
            and self.space_charge_mode.lower() == "2d"
        ):
            output += "Spacecharge3Dmesh();\n"
        else:
            output = ""
        return output


class gpt_tout(gpt_element):
    """
    Class for setting up the beam dump rate via `tout`.
    """

    startpos: float = 0.0
    """Starting position"""

    endpos: float = 0.0
    """End position"""

    starttime: float = 0.0
    """Start time for dumping"""

    step: str = "0.1/c"
    """Dump step as a string [distance / c]"""

    objectname: str = "tout"
    """Name of object"""

    objecttype: str = "gpt_tout"
    """Type of object"""


    def write_GPT(self, *args, **kwargs) -> str:
        self.starttime = 0 if self.starttime < 0 else self.starttime
        output = str(self.objectname) + "("
        if self.starttime is not None:
            output += str(self.starttime) + ","
        else:
            output += str(self.startpos) + "/c,"
        output += str(self.endpos) + ","
        output += str(self.step) + ");\n"
        return output


class gpt_csr1d(gpt_element):
    """
    Class for preparing CSR calculations via `csr1d`.
    """

    objectname: str = "csr1d"
    """Name of object"""

    objecttype: str = "gpt_csr1d"
    """Type of object"""

    def write_GPT(self, *args, **kwargs) -> str:
        output = str(self.objectname) + "();\n"
        return output


class gpt_writefloorplan(gpt_element):
    """
    Class for writing the lattice floor plan via `writefloorplan`.
    """

    filename: str = ""
    """Floor plan filename"""

    objectname: str = "writefloorplan"
    """Name of object"""

    objecttype: str = "gpt_writefloorplan"
    """Type of object"""


    def write_GPT(self, *args, **kwargs) -> str:
        output = str(self.objectname) + "(" + self.filename + ");\n"
        return output


class gpt_Zminmax(gpt_element):
    """
    Class for setting the boundaries in z for discarding particles via `Zminmax`
    """

    zmin: float = 0.0
    """Minimum longitudinal position"""

    zmax: float = 0.0
    """Maximum longitudinal position"""

    ECS: str = '"wcs", "I"'
    """Element coordinate system as a string"""

    objectname: str = "Zminmax"
    """Name of object"""

    objecttype: str = "gpt_Zminmax"
    """Type of object"""


    def write_GPT(self, *args, **kwargs):
        output = (
            str(self.objectname)
            + "("
            + self.ECS
            + ", "
            + str(self.zmin)
            + ", "
            + str(self.zmax)
            + ");\n"
        )
        return output


class gpt_forwardscatter(gpt_element):
    """
    Class for scattering particles via `forwardscatter`.
    """

    zmin: float = 0.0
    """Minimum longitudinal position"""

    zmax: float = 0.0
    """Maximum longitudinal position"""

    ECS: str = '"wcs", "I"'
    """Element coordinate system"""

    probability: float = 0.0
    """Scattering probability"""

    objectname: str = "forwardscatter"
    """Name of object"""

    objecttype: str = "gpt_forwardscatter"
    """Type of object"""


    def write_GPT(self, *args, **kwargs) -> str:
        output = (
            str(self.objectname)
            + "("
            + self.ECS
            + ', "'
            + str(self.name)
            + '", '
            + str(self.probability)
            + ");\n"
        )
        return output


class gpt_scatterplate(gpt_element):
    """
    Class for scattering particles off a plate via `scatterplate`.
    """

    zmin: float = 0.0
    """Minimum longitudinal position"""

    zmax: float = 0.0
    """Maximum longitudinal position"""

    ECS: str = '"wcs", "I"'
    """Element coordinate system"""

    a: float = 0.0
    """Plate length in x-direction"""

    b: float = 0.0
    """Plate length in y-direction"""

    model: str = "cathode"
    """Scattering model to be used ['cathode', 'remove']"""

    objectname: str = "scatterplate"
    """Name of object"""

    objecttype: str = "gpt_scatterplate"
    """Type of object"""


    def write_GPT(self, *args, **kwargs) -> str:
        output = (
            str(self.objectname)
            + "("
            + self.ECS
            + ", "
            + str(self.a)
            + ", "
            + str(self.b)
            + ') scatter="'
            + str(self.model)
            + '";\n'
        )
        return output


class gpt_dtmaxt(gpt_element):
    """
    Class for setting up minimum, maximmum temporal step sizes for tracking via `dtmaxt`.
    """

    tend: float = 0.0
    """Final time value"""

    tstart: float = 0.0
    """Initial time value"""

    dtmax: float = 0.0
    """Maximum temporal step size"""

    objectname: str = "dtmaxt"
    """Name of object"""

    objecttype: str = "gpt_dtmaxt"
    """Type of object"""


    def write_GPT(self, *args, **kwargs) -> str:
        output = (
            str(self.objectname)
            + "("
            + str(self.tstart)
            + ", "
            + str(self.tend)
            + ", "
            + str(self.dtmax)
            + ");\n"
        )
        return output


class gpt_dtmint(gpt_element):
    """
    Class for specifying the minimum timestep size via `dtmin`.
    """

    dtmin: float = 0.0
    """Maximum temporal step size"""

    objectname: str = "dtmin"
    """Name of object"""

    objecttype: str = "gpt_dtmint"
    """Type of object"""

    def write_GPT(self, *args, **kwargs):
        output = (
            str(self.objectname)
            + " = "
            + str(self.dtmin)
            + ";\n"
        )
        return output



import os
import numpy as np
import yaml
from pydantic import BaseModel, ConfigDict
from pydantic_core import PydanticUndefinedType
from typing import Dict, List
import xtrack as xt
from xtrack.beam_elements.elements import _HasKnlKsl, Bend
from nala.models.elementList import (
    SectionLattice,
    MachineLayout,
    ElementList,
)
from . import magnetic_orders
from ...utils.functions import introspect_model_defaults
from ...conversion_rules.codes import xsuite_conversion
from warnings import warn

type_conversion_rules_xsuite_reversed = xsuite_conversion.xsuite_conversion_rules_reverse


with open(
    os.path.dirname(os.path.abspath(__file__)) +
    "/../../conversion_rules/keywords/keyword_conversion_rules_Xsuite.yaml",
    "r",
) as infile:
    keyword_conversion_rules = yaml.safe_load(infile)

class XsuiteLatticeConverter(BaseModel):

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
        validate_assignment=True,
    )

    name: str = "Lattice"

    machine_area: str = "Lattice"

    line: xt.Line
    """Xsuite line"""

    elements: Dict = {}
    """Dictionary containing converted element objects"""

    sections: Dict = {}

    layouts: Dict = {}

    def rotation_matrix_from_survey(self, row):
        """Builds rotation matrix (local→global) from survey row."""
        R = np.array([
            [row["ex"][0], row["ex"][1], row["ex"][2]],
            [row["ey"][0], row["ey"][1], row["ey"][2]],
            [row["ez"][0], row["ez"][1], row["ez"][2]],
        ])
        return R

    def Ry(self, angle):
        """Rotation matrix about local y axis."""
        c, s = np.cos(angle), np.sin(angle)
        return np.array([[c, 0, s],
                         [0, 1, 0],
                         [-s, 0, c]])

    def compute_element_center(self, P0, R0, L, theta=0.0):
        """
        Given start position/orientation, return global center position and orientation.
        theta is total bending angle (radians); positive = bend right (horizontal plane).
        """
        # --- position of midpoint in local coordinates ---
        if abs(theta) < 1e-12:  # straight
            local_mid = np.array([0.0, 0.0, L / 2])
        else:
            Rbend = L / theta
            phi = theta / 2
            local_mid = np.array([
                Rbend * (1 - np.cos(phi)),  # x
                0.0,  # y
                Rbend * np.sin(phi)  # s
            ])

        # --- global midpoint ---
        Pcenter = P0 + R0 @ local_mid

        # --- global rotation at center ---
        if abs(theta) < 1e-12:
            Rcenter = R0
        else:
            Rcenter = R0 @ self.Ry(theta / 2)

        return Pcenter, Rcenter

    # Example for a batch of elements:
    def midpoints_for_line(self, element_and_survey,
                           local_axes_map=None):
        """
        Compute midpoints for many elements.

        Inputs
        - survey_positions: (N,3) array of start positions P0 for each element
        - survey_rotations: (N,3,3) array of rotations R0 for each element
          (R0 maps local->global)
        - elements: sequence of element-objects or dicts, each must provide:
            - length: float
            - angle: float   # bending angle in radians; if absent treated as 0
          Accepts any object where `getattr(el,'length')` and `getattr(el,'angle',0.0)` work,
          or dict-like with keys 'length' and 'angle'.
        - local_axes_map: see element_midpoint_global

        Returns
        - mids: (N,3) ndarray of midpoint positions
        """
        elem_pos = {}
        yhat = np.array([0, 1, 0])  # assuming horizontal bending plane

        for i, survey in enumerate(element_and_survey.values()):
            el = self.line.elements[i]
            # try several common attribute names for angle (xtrack names vary)
            L = getattr(el, 'length', getattr(el, 'L', 0.0))
            theta = getattr(el, 'angle', getattr(el, 'bending_angle', 0.0))
            P0 = np.array([survey["x"], survey["y"], survey["z"]])
            R0 = self.rotation_matrix_from_survey(survey)

            Pmid, Rmid = self.compute_element_center(P0, R0, L, theta)
            ex, ey, ez = Rmid[:, 0], Rmid[:, 1], Rmid[:, 2]

            elem_pos.update(
                {
                    list(element_and_survey.keys())[i]: {
                        "x": Pmid[0],
                        "y": Pmid[1],
                        "z": Pmid[2],
                        "phi": survey["phi"],
                        "psi": survey["psi"],
                        "theta": survey["theta"],
                    }
                }
            )
        return elem_pos

    def create_element_dictionary(self):
        s = self.line.survey()._data
        elems = {k: v for k, v in zip(s["name"], self.line.elements)}
        survey = {
            s["name"][i]: {
                "x": s["X"][i],
                "y": s["Y"][i],
                "z": s["Z"][i],
                "ex": s["ex"][i],
                "ey": s["ey"][i],
                "ez": s["ez"][i],
                "phi": (s["phi"][i] + np.pi) % (2*np.pi) - np.pi,
                "psi": (s["psi"][i] + np.pi) % (2*np.pi) - np.pi,
                "theta": (s["theta"][i] + np.pi) % (2*np.pi) - np.pi,
            } for i in range(len(s["X"][:-1]))
        }
        survey = self.midpoints_for_line(survey)
        for k, v in elems.items():
            machine_area = "IOTA_v8pt4_madx"
            length = 0
            if hasattr(v, "length"):
                length = v.length
            pos = [float(x) for x in [survey[k]["x"], survey[k]["y"], survey[k]["z"]]]
            rot = [float(x) for x in [survey[k]["phi"], survey[k]["psi"], survey[k]["theta"]]]
            phys = {"middle": pos, "global_rotation": rot, "length": length}

            if type(v) in type_conversion_rules_xsuite_reversed:
                p_obj = type_conversion_rules_xsuite_reversed[type(v)]
                model_fields = introspect_model_defaults(p_obj)
                hardware_class = p_obj.model_fields["hardware_class"].default
                if not type(p_obj.model_fields["hardware_type"].default) == PydanticUndefinedType:
                    hardware_type = p_obj.model_fields["hardware_type"].default
                else:
                    hardware_type = hardware_class
                if type(hardware_class) == PydanticUndefinedType:
                    hardware_class = hardware_type
                newobj = {
                    "name": k,
                    "hardware_type": hardware_type,
                    "hardware_class": hardware_class,
                    "machine_area": machine_area,
                    "physical": phys,
                }
                try:
                    merged = keyword_conversion_rules[hardware_type.lower()] | keyword_conversion_rules["general"]
                except KeyError:
                    merged = keyword_conversion_rules["general"]
                except TypeError:
                    merged = keyword_conversion_rules["general"]
                for subk in ["magnetic", "cavity", "simulation", "diagnostic"]:
                    if subk in model_fields:
                        newobj.update({subk: {}})
                kwele = {y: x for x, y in merged.items()}
                exclude = ["order"]
                for name in dir(v):
                    for subk in model_fields:
                        if isinstance(model_fields[subk], dict) and name not in exclude:
                            if name in ["k1", "k2", "k3", "angle"] and isinstance(v, _HasKnlKsl):
                                if "magnetic" not in newobj:
                                    newobj.update({"magnetic": {"length": length}})
                                try:
                                    newobj["magnetic"]["kl"] = getattr(
                                        v, f"k{magnetic_orders[newobj['hardware_type']]}"
                                    ) * v.length
                                except AttributeError as e:
                                    print(e)
                                    newobj["magnetic"]["kl"] = getattr(v, name)
                                newobj["magnetic"].update({name: getattr(v, name)})
                                if name == "angle":
                                    newobj["magnetic"]["kl"] = newobj["magnetic"]["kl"] * -1
                                newobj["hardware_class"] = "Magnet"
                            if name in ["ks"]:
                                newobj.update({"magnetic": {"ks": v.ks, "length": v.length}})
                            if name in model_fields[subk]:
                                newobj[subk].update({name: getattr(v, name)})
                            elif name in kwele:
                                if kwele[name] in model_fields[subk]:
                                    if not isinstance(model_fields[subk][kwele[name]], str) or model_fields[subk][
                                        kwele[name]]:
                                        try:
                                            newobj[subk].update({kwele[name]: getattr(v, name)})
                                        except KeyError as e:
                                            print(e)
                                        except AttributeError as e:
                                            print(e)
                if "angle" in newobj["physical"]:
                    newobj["physical"].update({"angle": newobj["physical"]["angle"] * -1})
                # if not newobj["hardware_class"] == "Drift":
                self.elements.update({k: p_obj(**newobj)})
            else:
                warn(f"Type conversion {type(v)} not implemented")

    def create_section(self, start: str, end: str, name: str) -> Dict[str, SectionLattice]:
        if not self.elements:
            self.create_element_dictionary()
        appending = False
        order = []
        elems = {}
        if any([start not in self.elements.keys(), end not in self.elements.keys()]):
            raise KeyError(f"Could not find {start} or {end} in lattice; please check names")
        for name, elem in self.elements.items():
            if name == start:
                appending = True
            if appending:
                order.append(name)
                elems.update({name: elem})
            if name == end:
                appending = False
        if not elems:
            raise ValueError(f"Could not create list of elements; check {start} is before {end}")
        seclat = SectionLattice(order=order, elements=ElementList(elements=elems), name=name)
        self.sections.update({name: seclat})
        return {name: seclat}

    def create_layout(self, name: str, sections: List[Dict]) -> MachineLayout:
        if not self.elements:
            self.create_element_dictionary()
        layout_sections = {}
        for sec in sections:
            assert "name" in sec and "start" in sec and "end" in sec
            layout_sections.update(
                self.create_section(
                    start=sec["start"],
                    end=sec["end"],
                    name=sec["name"],
                )
            )
        layout = MachineLayout(
            name=name,
            sections={k: v for k, v in zip(list(layout_sections.keys()), list(layout_sections.values()))}
        )
        self.layouts.update({name: layout})
        return layout


from pydantic import Field, computed_field
from typing import Type, Union, Literal
from warnings import warn
from scipy.constants import pi, c, e, m_e, epsilon_0
import numpy as np

from .baseModels import IgnoreExtra, T


class LaserElement(IgnoreExtra):
    """Laser info model."""

    initial_position: float = 0
    """Initial position of the laser pulse [m]"""

    waist: float = Field(ge=0, default=0)
    """Laser waist [m]"""

    wavelength: float = Field(gt=0)
    """Laser wavelength [m]"""

    pulse_energy: float = Field(gt=0)
    """Laser pulse energy [J]"""

    pulse_duration_fwhm: float = Field(gt=0)
    """Pulse duration FWHM [s]"""

    focal_position: float = 0.0
    """Focal position of the laser pulse [m], optional"""

    cep_phase: float = 0
    """CEP phase [radians]"""

    polarization: Literal["linear", "circular", "elliptical"] | None = None
    """Laser polarization: 'linear', 'circular', 'elliptical'"""

    profile_type: Literal['gaussian', 'laguerre-gaussian', 'flattened-gaussian', 'file'] = "gaussian"
    """Laser profile type [str]: 'gaussian', 'laguerre-gaussian', 'flattened-gaussian', 'file'"""

    laguerre_polynomial_order_p: int = 0
    """Order of Laguerre-Gaussian polynomial mode, if profile_type is 'laguerre-gaussian'"""

    flatness: int = 6
    """Flatness parameter, if profile_type is 'flattened-gaussian'.
    Default: N=6; somewhat close to an 8th order super-gaussian."""

    @computed_field
    @property
    def amplitude(self) -> float:
        """
        Laser amplitude: ((e*lambda0)/(pi*m_e*c**2*w0)) * np.sqrt( E/(pi*epsilon_0*c*tau_FWHM) )

        Returns
        -------
        float
            Laser amplitude (dimensionless)

        Raises
        ------
        ValueError
            If any of the requires parameters are not set or non-positive
        """
        if any([self.wavelength, self.waist, self.pulse_energy, self.pulse_duration_fwhm]) <= 0:
            warn("Wavelength, waist, pulse enegy and pulse duration must be positive "
                 "to compute laser amplitude.")
            return 0
        return ((e * self.wavelength) / (pi * m_e * c ** 2 * self.waist)) * np.sqrt(
            self.pulse_energy / (pi * epsilon_0 * c * self.pulse_duration_fwhm)
        )

    @property
    def angular_frequency(self) -> float:
        """
        Laser angular frequency: 2*pi*c/lambda0

        Returns
        -------
        float
            Laser angular frequency [rad/s]

        Raises
        ------
        ValueError
            If wavelength is not set or non-positive
        """
        if self.wavelength <= 0:
            raise ValueError("Wavelength must be positive to compute laser angular frequency.")
        return 2 * pi * c / self.wavelength


class LaserHalfWavePlateElement(IgnoreExtra):
    """
    Laser half-wave plate model.
    """

    calibration_factor: float = Field()
    """Calibration factor for the half-wave plate."""

    pv_type: str = Field(alias="laser_pv_type")
    """Type of the laser PV."""

class LaserEnergyMeterElement(IgnoreExtra):
    """
    Laser energy meter model.
    """

    calibration_factor: float = Field()
    """Calibration factor for the energy meter, i.e. between measured value and actual laser energy."""

    pv_type: str = Field(alias="laser_pv_type")
    """Type of the laser PV."""


class LaserMirrorSense(IgnoreExtra):
    """
    Laser mirror sense model.
    """

    left: float = Field(alias="left_sense")
    right: float = Field(alias="right_sense")
    up: float = Field(alias="up_sense")
    down: float = Field(alias="down_sense")


class LaserMirrorElement(IgnoreExtra):
    """Laser info model."""

    step_max: float = Field()
    sense: LaserMirrorSense
    vertical_channel: Union[int, None] = None
    horizontal_channel: Union[int, None] = None

    @classmethod
    def from_CATAP(cls: Type[T], fields: dict) -> T:
        cls._create_field_class(cls, fields, "sense", LaserMirrorSense)
        return super().from_CATAP(fields)



"""
NALA Main Module

The main class for handling a full particle accelerator lattice.
"""
import os
import glob
from math import copysign
from itertools import chain
from typing import List, Dict
from pydantic import field_validator
from yaml.constructor import Constructor

from .models.physical import PhysicalElement, Position
from .models.elementList import MachineModel, baseElement
from .models.element import Drift
from .Importers.YAML_Loader import (
    read_YAML_Combined_File,
    read_YAML_Element_Files,
    interpret_YAML_Element,
)
import numpy as np


def flatten(xss):
    """Flatten a list of lists."""
    return list(chain.from_iterable(xss))


def add_bool(self, node):
    return self.construct_scalar(node)


Constructor.add_constructor("tag:yaml.org,2002:bool", add_bool)


def dot(a, b) -> float:
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


def chunks(li, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(li), n):
        yield li[i: i + n]


class NALA(MachineModel):
    """
    NALA Main Class

    The main class for handling a full particle accelerator lattice.
    """

    element_list: str | List[baseElement]
    """List containing all elements in the machine model, either as a path to a YAML file/directory 
    or as a list of element objects."""

    @field_validator("element_list", mode="before")
    @classmethod
    def validate_element_list(cls, v: str | list) -> str | list:
        if isinstance(v, str):
            if os.path.isfile(v):
                return v
            elif os.path.isfile(os.path.abspath(os.path.dirname(__file__) + "/" + v)):
                return os.path.abspath(os.path.dirname(__file__) + "/" + v)
            elif os.path.isdir(v):
                return v
            elif os.path.isdir(os.path.abspath(os.path.dirname(__file__) + "/" + v)):
                return os.path.abspath(os.path.dirname(__file__) + "/" + v)
            else:
                raise ValueError(f"Directory {v} does not exist")
        else:
            return v

    def model_post_init(self, __context):
        super().model_post_init(__context)
        if isinstance(self.element_list, str):
            if os.path.isfile(self.element_list):
                elems = read_YAML_Combined_File(self.element_list)
            elif os.path.isdir(self.element_list):
                files = glob.glob(
                    os.path.abspath(self.element_list + "/**/*.yaml"), recursive=True
                )
                data = read_YAML_Element_Files(files)
                elems = [interpret_YAML_Element(data) for data in data]
        else:
            elems = self.element_list
        self.update({y.name: y for y in elems})

    def createDrifts(self, end: str = None, start: str = None, path: str = None) -> Dict:
        """
        Insert drifts into a sequence of 'elements'

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: Dictionary of elements with drifts inserted
        """
        positions = []
        originalelements = dict()
        elementno = 0
        newelements = dict()

        elements = self.elements_between(
            start=start, end=end, element_class=None, path=path
        )

        for name in elements:
            elem = self.elements[name]
            originalelements[name] = elem
            pos = elem.physical.start.array
            positions.append(pos)
            positions.append(elem.physical.end.array)
        positions = positions[1:]
        positions.append(positions[-1])
        driftdata = list(
            zip(iter(list(originalelements.items())), list(chunks(positions, 2)))
        )

        for e, d in driftdata:
            newelements[e[0]] = e[1]
            if len(d) > 1:
                x1, y1, z1 = d[0]
                x2, y2, z2 = d[1]
                try:
                    length = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2)
                    vector = dot((d[1] - d[0]), [0, 0, 1])
                except Exception as exc:
                    print("Element with error = ", e[0])
                    print(d)
                    raise exc
                if round(length, 6) > 0:
                    elementno += 1
                    name = "drift" + str(elementno)
                    x, y, z = [(a + b) / 2.0 for a, b in zip(d[0], d[1])]
                    newdrift = Drift(
                        name=name,
                        machine_area=newelements[e[0]].machine_area,
                        hardware_class="drift",
                        physical=PhysicalElement(
                            length=round(copysign(length, vector), 6),
                            middle=Position(x=x, y=y, z=z),
                            datum=Position(x=x, y=y, z=z),
                        ),
                    )
                    newelements[name] = newdrift
        return newelements

    def get_elements(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all elements between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of element names
        """
        return self.elements_between(
            start=start, end=end, element_class=None, path=path
        )

    def _drift_length(self, start: list[float], end: list[float]):
        return np.linalg.norm(end - start)

    def get_elements_s_pos(self, end: str = None, start: str = None, path: str = None) -> Dict[str, float]:
        """
        Get s positions of all elements between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: Dictionary of element names and their s positions
        """
        elements = self.createDrifts(start=start, end=end, path=path)
        start_and_end = [
            [name, elem.physical.length, elem.hardware_type == "Drift"]
            for name, elem in elements.items()
        ]
        elem_s = {}
        s_pos = 0
        for elem, l, drift in start_and_end:
            s_pos += l
            if not drift:
                elem_s[elem] = round(s_pos, 6)
        return elem_s

    def get_rf_cavities(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all RF cavities between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of RF cavity names
        """
        return self.elements_between(
            start=start, end=end, element_class="rf", path=path
        )

    def get_diagnostics(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all diagnostic devices between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of diagnostic devices names
        """
        return self.elements_between(
            start=start, end=end, element_class="diagnostic", path=path
        )

    def get_charge_diagnostics(
        self, end: str = None, start: str = None, path: str = None
    ) -> list[str]:
        """
        Get all charge diagnostic devices between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of charge diagnostic devices names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="diagnostic",
            element_type=["FCM", "WCM", "ICT"],
            path=path,
        )

    def get_beam_position_monitors(
        self, end: str = None, start: str = None, path: str = None
    ) -> list[str]:
        """
        Get all BPM devices between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of BPM devices names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="diagnostic",
            element_type="BPM",
            path=path,
        )

    def get_position_diagnostics(
        self, end: str = None, start: str = None, path: str = None
    ) -> list[str]:
        """
        Get all position diagnostic devices between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of position diagnostic devices names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="diagnostic",
            element_type=["Screen", "BPM"],
            path=path,
        )

    def get_cameras(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all camera devices between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of camera devices names
        """
        return [
            self[scr].diagnostic.camera_name
            for scr in self.elements_between(
                start=start,
                end=end,
                element_class="diagnostic",
                element_type="Screen",
                path=path,
            )
        ]

    def get_screens_and_cameras(
        self, end: str = None, start: str = None, path: str = None
    ) -> Dict[str, str]:
        """
        Get all screen devices with their associated cameras between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: Dict of screens with camera names
        """
        return {
            scr: self[scr].diagnostic
            for scr in self.elements_between(
                start=start,
                end=end,
                element_class="diagnostic",
                element_type="Screen",
                path=path,
            )
        }

    def get_magnets(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of magnet names
        """
        return self.elements_between(
            start=start, end=end, element_class="magnet", path=path
        )

    def get_separate_magnets(
        self, end: str = None, start: str = None, path: str = None
    ) -> list[str]:
        """
        Get all magnet between start and end and separate combined correctors into their sub-correctors

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of magnet names
        """
        magnets = self.get_magnets(end=end, start=start, path=path)
        return list(
            flatten(
                [(self.__get_combined_corrector_sub_correctors(c)) for c in magnets]
            )
        )

    def get_quadrupoles(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all quadrupole magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of quadrupole magnet names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="magnet",
            element_type="quadrupole",
            path=path,
        )

    def get_dipoles(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all dipole magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of dipole magnet names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="magnet",
            element_type="dipole",
            path=path,
        )

    def __get_combined_corrector_sub_correctors(self, elem: str) -> list[str]:
        """
        Split a Combined_Corrector into its sub-correctors

        :param elem: Name of the combined corrector element
        :return: Names of sub-corrector elements (if they exist) or the original element name
        """
        if (
            hasattr(self[elem], "Horizontal_Corrector")
            and self[elem].Horizontal_Corrector is not None
        ):
            if (
                hasattr(self[elem], "Vertical_Corrector")
                and self[elem].Vertical_Corrector is not None
            ):
                return [self[elem].Horizontal_Corrector, self[elem].Vertical_Corrector]
            else:
                return [self[elem].Horizontal_Corrector]
        elif (
            hasattr(self[elem], "Vertical_Corrector")
            and self[elem].Vertical_Corrector is not None
        ):
            return [self[elem].Vertical_Corrector]
        else:
            return [elem]

    def get_correctors(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all corrector magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of corrector magnet names
        """
        correctors = self.elements_between(
            start=start,
            end=end,
            element_class="magnet",
            element_type=[
                "combined_corrector",
                "horizontal_corrector",
                "vertical_corrector",
            ],
            path=path,
        )
        return list(
            flatten(
                [(self.__get_combined_corrector_sub_correctors(c)) for c in correctors]
            )
        )

    def get_horizontal_correctors(
        self, end: str = None, start: str = None, path: str = None
    ) -> list[str]:
        """
        Get all horizontal corrector magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of horizontal corrector magnet names
        """
        horizontal_correctors = self.elements_between(
            start=start,
            end=end,
            element_class="magnet",
            element_type=["combined_corrector", "horizontal_corrector"],
            path=path,
        )
        return [
            (
                self[c].Horizontal_Corrector
                if hasattr(self[c], "Horizontal_Corrector")
                and self[c].Horizontal_Corrector is not None
                else c
            )
            for c in horizontal_correctors
        ]

    def get_vertical_correctors(
        self, end: str = None, start: str = None, path: str = None
    ) -> list[str]:
        """
        Get all vertical corrector magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of vertical corrector magnet names
        """
        vertical_correctors = self.elements_between(
            start=start,
            end=end,
            element_class="magnet",
            element_type=["combined_corrector", "Vertical_Corrector"],
            path=path,
        )
        return [
            (
                self[c].Vertical_Corrector
                if hasattr(self[c], "Vertical_Corrector")
                and self[c].Vertical_Corrector is not None
                else c
            )
            for c in vertical_correctors
        ]

    def get_lattice_correctors(
        self, end: str = None, start: str = None, path: str = None
    ) -> list[str]:
        """
        Get all corrector magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of corrector magnet names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="magnet",
            element_type=[
                "combined_corrector",
                "horizontal_corrector",
                "vertical_corrector",
            ],
            path=path,
        )

    def get_combined_correctors(
        self, end: str = None, start: str = None, path: str = None
    ) -> list[str]:
        """
        Get all combined corrector magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of combined corrector magnet names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="magnet",
            element_type=["combined_corrector"],
            path=path,
        )

    def get_sextupoles(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all sectupole magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of sectupole magnet names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="magnet",
            element_type="sextupole",
            path=path,
        )

    def get_solenoids(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all solenoid magnets between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of solenoid magnet names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="magnet",
            element_type="solenoid",
            path=path,
        )

    def get_vacuum_components(
        self, end: str = None, start: str = None, path: str = None
    ) -> list[str]:
        """
        Get all vacuum components between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of vacuum component names
        """
        return self.elements_between(
            start=start, end=end, element_class="vacuum", path=path
        )

    def get_shutters(self, end: str = None, start: str = None, path: str = None) -> list[str]:
        """
        Get all shutter devices between start and end

        :param end: Name of the last element in the sequence
        :param start: Name of the first element in the sequence
        :param path: Name of the lattice path to use
        :return: List of shutter devices names
        """
        return self.elements_between(
            start=start,
            end=end,
            element_class="vacuum",
            element_type="shutter",
            path=path,
        )

    def __all_elements(
        self, element_class: str | list | None = None, element_type: str | list | None = None
    ) -> set:
        """
        Get a set of all elements of a given class and/or type

        :param element_class: Class of the element (e.g., 'magnet', 'rf', 'diagnostic')
        :param element_type: Type of the element (e.g., 'quadrupole', 'dipole', 'BPM')
        :return: Set of element names
        """
        return set(
            [
                elem
                for pathelems in [
                    self.elements_between(
                        start=None,
                        end=None,
                        element_class=element_class,
                        element_type=element_type,
                        path=path,
                    )
                    for path in self.lattices.keys()
                ]
                for elem in pathelems
            ]
        )

    @property
    def all_elements(self) -> set:
        """
        Get all elements in the machine
        :return: Set of all element names
        """
        return self.__all_elements()

    @property
    def all_rf_cavities(self) -> set:
        """
        Get all rf cavities in the machine
        :return: Set of all cavity names
        """
        return self.__all_elements(element_class="rf")

    @property
    def all_diagnostics(self) -> set:
        """
        Get all diagnostic devices in the machine
        :return: Set of all diagnostic device names
        """
        return self.__all_elements(element_class="diagnostic")

    @property
    def all_charge_diagnostics(self) -> set:
        """
        Get all charge diagnostics in the machine
        :return: Set of all charge diagnostic names
        """
        return self.__all_elements(
            element_class="diagnostic",
            element_type=["FCM", "WCM", "ICT"],
        )

    @property
    def all_beam_position_monitors(self) -> set:
        """
        Get all BPM devices in the machine
        :return: Set of all BPM names
        """
        return self.__all_elements(
            element_class="diagnostic",
            element_type="BPM",
        )

    @property
    def all_position_diagnostics(self) -> set:
        """
        Get all position diagnostic devices in the machine
        :return: Set of all position diagnostic device names
        """
        return self.__all_elements(
            element_class="diagnostic",
            element_type=["Screen", "BPM"],
        )

    @property
    def all_cameras(self) -> list:
        """
        Get all camera devices in the machine
        :return: List of all camera device names
        """
        return [
            self[scr].diagnostic.camera_name
            for scr in self.__all_elements(
                element_class="diagnostic",
                element_type="Screen",
            )
        ]

    @property
    def all_screens_and_cameras(self) -> Dict:
        """
        Get all screens with their associated cameras in the machine
        :return: Dict of all screens with camera names
        """
        return {
            scr: self[scr].diagnostic.camera_name
            for scr in self.__all_elements(
                element_class="diagnostic",
                element_type="Screen",
            )
        }

    @property
    def all_magnets(self) -> set:
        """
        Get all magnets in the machine
        :return: Set of all magnet names
        """
        return self.__all_elements(element_class="magnet")

    @property
    def all_quadrupoles(self) -> set:
        """
        Get all quadrupole magnets in the machine
        :return: Set of all quadrupole magnet names
        """
        return self.__all_elements(
            element_class="magnet",
            element_type="quadrupole",
        )

    @property
    def all_dipoles(self) -> set:
        """
        Get all dipole magnets in the machine
        :return: Set of all dipole magnet names
        """
        return self.__all_elements(
            element_class="magnet",
            element_type="dipole",
        )

    @property
    def all_combined_correctors(self) -> set:
        """
        Get all combined corrector magnets in the machine
        :return: Set of all combined corrector magnet names
        """
        return self.__all_elements(
            element_class="magnet",
            element_type="combined_corrector",
        )

    @property
    def all_separate_magnets(self) -> set:
        """
        Get all separate magnets in the machine
        :return: Set of all separate magnet names
        """
        return set(
            [
                elem
                for pathelems in [
                    self.get_separate_magnets(start=None, end=None, path=path)
                    for path in self.lattices.keys()
                ]
                for elem in pathelems
            ]
        )

    @property
    def all_correctors(self) -> set:
        """
        Get all corrector magnets in the machine
        :return: Set of all corrector magnet names
        """
        return set(
            [
                elem
                for pathelems in [
                    self.get_correctors(
                        start=None,
                        end=None,
                        path=path,
                    )
                    for path in self.lattices.keys()
                ]
                for elem in pathelems
            ]
        )

    @property
    def all_horizontal_correctors(self) -> set:
        """
        Get all horizontal corrector magnets in the machine
        :return: Set of all horizontal corrector magnet names
        """
        return set(
            [
                elem
                for pathelems in [
                    self.get_horizontal_correctors(
                        start=None,
                        end=None,
                        path=path,
                    )
                    for path in self.lattices.keys()
                ]
                for elem in pathelems
            ]
        )

    @property
    def all_vertical_correctors(self) -> set:
        """
        Get all vertical corrector magnets in the machine
        :return: Set of all vertical corrector magnet names
        """
        return set(
            [
                elem
                for pathelems in [
                    self.get_vertical_correctors(
                        start=None,
                        end=None,
                        path=path,
                    )
                    for path in self.lattices.keys()
                ]
                for elem in pathelems
            ]
        )

    @property
    def all_sextupoles(self) -> set:
        """
        Get all sextupole magnets in the machine
        :return: Set of all sextupole magnet names
        """
        return self.__all_elements(
            element_class="magnet",
            element_type="sextupole",
        )

    @property
    def all_solenoids(self) -> set:
        """
        Get all solenoid magnets in the machine
        :return: Set of all solenoid magnet names
        """
        return self.__all_elements(
            element_class="magnet",
            element_type="solenoid",
        )

    @property
    def all_vacuum_components(self) -> set:
        """
        Get all vacuum components in the machine
        :return: Set of all vacuum component names
        """
        return self.__all_elements(element_class="vacuum")

    @property
    def all_shutters(self) -> set:
        """
        Get all shutter elements in the machine
        :return: Set of all shutter element names
        """
        return self.__all_elements(
            element_class="vacuum",
            element_type="shutter",
        )



from pydantic import field_validator
from .baseModels import IgnoreExtra


class ManufacturerElement(IgnoreExtra):
    """Manufacturer info model."""

    manufacturer: str = ""
    """Name of manufacturer."""

    serial_number: str = ""
    """Serial number of element."""

    @field_validator("serial_number", mode="before")
    @classmethod
    def validate_serial_number(cls, v: str | int) -> str:
        if isinstance(v, int):
            return str(v)
        return v

    @field_validator("manufacturer", mode="before")
    @classmethod
    def validate_manufacturer(cls, v: str | int) -> str:
        if isinstance(v, int):
            return str(v)
        return v


import os
import yaml

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/types/type_conversion_rules.yaml",
    "r",
) as infile:
    type_conversion_rules = yaml.safe_load(infile)
    type_conversion_rules_Elegant = type_conversion_rules["elegant"]
    type_conversion_rules_Genesis = type_conversion_rules["genesis"]
    type_conversion_rules_Opal = type_conversion_rules["opal"]
    type_conversion_rules_Names = type_conversion_rules["name"]
    type_conversion_rules_aliases = type_conversion_rules["aliases"]["elegant"]

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/keywords/keyword_conversion_rules_elegant.yaml",
    "r",
) as infile:
    keyword_conversion_rules_elegant = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/elements/elements_elegant.yaml",
    "r",
) as infile:
    elements_Elegant = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/keywords/keyword_conversion_rules_ocelot.yaml",
    "r",
) as infile:
    keyword_conversion_rules_ocelot = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/elements/elements_ocelot.yaml",
    "r",
) as infile:
    elements_Ocelot = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/keywords/keyword_conversion_rules_cheetah.yaml",
    "r",
) as infile:
    keyword_conversion_rules_cheetah = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/elements/elements_cheetah.yaml",
    "r",
) as infile:
    elements_Cheetah = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/elements/elements_opal.yaml",
    "r",
) as infile:
    elements_Opal = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/keywords/keyword_conversion_rules_opal.yaml",
    "r",
) as infile:
    keyword_conversion_rules_opal = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/keywords/keyword_conversion_rules_Xsuite.yaml",
    "r",
) as infile:
    keyword_conversion_rules_xsuite = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/keywords/keyword_conversion_rules_wake_t.yaml",
    "r",
) as infile:
    keyword_conversion_rules_wake_t = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/keywords/keyword_conversion_rules_genesis.yaml",
    "r",
) as infile:
    keyword_conversion_rules_genesis = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/elements/elements_genesis.yaml",
    "r",
) as infile:
    elements_Genesis = yaml.safe_load(infile)

with open(
    os.path.dirname(os.path.abspath(__file__)) + "/../conversion_rules/elements/element_keywords.yaml",
    "r",
) as infile:
    element_keywords = yaml.safe_load(infile)


from pydantic import PositiveInt, PositiveFloat, SerializeAsAny
from typing import Literal, Any, Dict
from .baseModels import IgnoreExtra


class ApertureElement(IgnoreExtra):
    """Physical info model."""

    number_of_elements: int | None = None
    """Number of aperture elements"""

    horizontal_size: PositiveFloat = 1.0
    """Horizontal aperture size [m]"""

    vertical_size: PositiveFloat = 1.0
    """Vertical aperture size [m]"""

    shape: Literal["elliptical", "planar", "circular", "rectangular", "scraper"] | None = None
    """Aperture shape"""

    radius: float | None = None
    """Radius of aperture"""

    negative_extent: float | None = None
    """Longitudinal start position of an aperture"""

    positive_extent: float | None = None
    """Longitudinal end position of an aperture"""


class SimulationElement(IgnoreExtra):
    """
    Simulation element model.
    """

    field_definition: SerializeAsAny[Any] = None
    """String pointing to field definition"""

    wakefield_definition: SerializeAsAny[Any] = None
    """String pointing to wakefield definition"""

    field_reference_position: Literal["start", "middle", "end"] | None = None
    """Reference position for field file"""

    scale_field: int | float | bool = False
    """Flag indicating whether to scale the field from the field file"""


class MagnetSimulationElement(SimulationElement):
    """
    Magnet simulation element model.
    """

    n_kicks: int = 4
    """Number of kicks for tracking through the quad"""

    smooth: int | float | None = 2
    """Number of points to smooth the field map [ASTRA only]"""

    edge_field_integral: float = 0.5
    """Edge field integral for fringes"""

    edge1_effects: int = 1
    """Flag to indicate whether entrance edge effects are included"""

    edge2_effects: int = 1
    """Flag to indicate whether exit edge effects are included"""

    sr_enable: bool = True
    """Flag to enable SR calculations"""

    integration_order: int = 4
    """Runge-Kutta integration order"""

    nonlinear: int = 1
    """Flag to indicate whether to perform nonlinear calculations"""

    smoothing_half_width: int = 1
    """Half-width for smoothing"""

    edge_order: int = 2
    """Matrix order for edges"""

    csr_bins: int = 100
    """Number of CSR bins"""

    deltaL: float = 0.0
    """Delta length"""

    csr_enable: bool = True
    """Flag to indicate whether CSR is enabled"""

    isr_enable: bool = True
    """Flag to indicate whether ISR is enabled"""

    field_amplitude: float = 0.0


class DriftSimulationElement(SimulationElement):
    """
    Drift simulation element model.
    """

    lsc_interpolate: int = 1
    """Flag to allow for interpolation of computed longitudinal space charge wake.
    See `Elegant manual LSC drift`_
    
    .. _Elegant manual LSC drift: https://ops.aps.anl.gov/manuals/elegant_latest/elegantsu168.html#x179-18000010.58
    """

    csr_enable: bool = True
    """Enable CSR drift calculations"""

    lsc_enable: bool = True
    """Enable LSC drift calculations"""

    use_stupakov: int = 1
    """Use Stupakov formula; see `Elegant manual LSC drift`_"""

    csrdz: PositiveFloat = 0.01
    """Step size for CSR calculations"""

    lsc_bins: PositiveInt = 20
    """Number of bins for LSC calculations"""

    lsc_high_frequency_cutoff_start: float | None = None
    """Spatial frequency at which smoothing filter begins. If not positive, no frequency filter smoothing is done. 
    See `Elegant manual LSC drift`_
    """

    lsc_high_frequency_cutoff_end: float | None = None
    """Spatial frequency at which smoothing filter is 0. See `Elegant manual LSC drift`_"""

    lsc_low_frequency_cutoff_start: float | None = None
    """Highest spatial frequency at which low-frequency cutoff filter is zero. See `Elegant manual LSC drift`_"""

    lsc_low_frequency_cutoff_end: float | None = None
    """Lowest spatial frequency at which low-frequency cutoff filter is 1. See `Elegant manual LSC drift`_"""


class DiagnosticSimulationElement(SimulationElement):
    output_filename: str | None = None
    """Output filename for the diagnostic"""


class PlasmaSimulationElement(SimulationElement):
    """
    Plasma simulation element model.
    """

    wakefield_model: Literal["quasistatic_2d"] | None = None
    """Wakefield model (Wake-T); possible values: 
    'blowout', 'custom_blowout', 'focusing_blowout', 'cold_fluid_1d' and 'quasistatic_2d'; if None, no
    wakefields are computed.
    # TODO add more of these and check which we support
    """

    required_attrs: Dict = {
        "common": [
            "length",
        ],
        "quasistatic_2d": [
            "density",
            "r_max",
            "n_longitudinal",
            "n_radial",
            "min_longitudinal_position",
            "max_longitudinal_position"
        ],
    }

    bunch_pusher: Literal["rk4", "boris"] = "boris"
    """Pusher used to evolve particles in time in the plasma [Wake-T]; possible values:
    'rk4', 'boris'; see `Wake-T pusher`_

    .. _Wake-T pusher: https://github.com/AngelFP/Wake-T/tree/dev/wake_t/particles/push
    """

    dt_bunch: float | Literal["auto"] = "auto"
    """The time step for evolving the particle bunches. If 'auto', set to dt=T/(10*2*pi) with T
    the plasma period.
    """

    n_out: int = 1
    """Number of times to dump the particle distribution during the plasma stage."""

    min_longitudinal_position: float = 0
    """Minimum longitudinal position [metres] up to which
    plasma wakefield will be calculated. Converted to boosted co-ordinates for Wake-T during 
    simulation setup."""

    max_longitudinal_position: float = 0
    """Maximum longitudinal position [metres] up to which
        plasma wakefield will be calculated. Converted to boosted co-ordinates for Wake-T during
        simulation setup."""

    n_longitudinal: int = 0
    """Number of grid points in the longitudinal direction"""

    n_radial: int = 0
    """Number of grid points in the radial direction"""

    plasma_particles_per_cell: int = 2
    """Number of plasma particles per cell; 2 by default"""

    r_max: float = 0
    """Radial extent of the simulation box [meters]"""

    r_max_plasma: float | None = None
    """Maximum radial extension of the plasma column; set to `r_max` if None"""

    dz_fields: float | None = None
    """Determines how often the plasma wakefields should be updated; 
        `max_longitudinal_position`-`min_longitudinal_position` by default"""

    plasma_pusher: Literal["rk4", "boris"] = "boris"
    """Pusher used to evolve the plasma in time [Wake-T]; possible values:
    'rk4', 'boris'; see `Wake-T pusher`_

    .. _Wake-T pusher: https://github.com/AngelFP/Wake-T/tree/dev/wake_t/particles/push
    """


class RFCavitySimulationElement(SimulationElement):
    """
    RF cavity simulation element model.
    """

    field_amplitude: float = 0
    """Cavity field amplitude"""

    t_column: str | None = None
    """t column in wake file"""

    z_column: str | None = None
    """z column in wake file"""

    wx_column: str | None = None
    """Wx column in wake file"""

    wy_column: str | None = None
    """Wy column in wake file"""

    wz_column: str | None = None
    """Wz column in wake file"""

    change_p0: int = 1
    """Flag to indicate whether cavity is changing momentum"""

    n_kicks: int = 0
    """Number of cavity kicks to apply"""

    end1_focus: int = 1
    """Apply entrance focusing"""

    end2_focus: int = 1
    """Apply exit focusing"""

    body_focus_model: str = "SRS"
    """Cavity focusing model"""

    lsc_bins: int = 100
    """Number of longitudinal space charge bins"""

    current_bins: int = 0
    """Number of current bins"""

    interpolate_current_bins: int = 1
    """Flag to indicate whether to interpolate during current histogram"""

    smooth_current_bins: int = 1
    """Flag to indicate whether to smooth the current histogram"""

    smooth: int | None = None
    """Smoothing parameter"""

    ez_peak: float | None = None
    """Peak longitudinal electric field"""

    field_file_name: str | None = None
    """Cavity field file name"""

    wakefile: str | None = None
    """Name of wake file"""

    zwakefile: str | None = None
    """Name of longitudinal wake file"""

    trwakefile: str | None = None
    """Name of transverse wake file"""


class WakefieldSimulationElement(SimulationElement):
    """
    Wakefield simulation element model.
    """

    allow_long_beam: bool = True
    """Flag to indicate whether beams longer than the wakefield are allowed."""

    bunched_beam: bool = False
    """Flag to indicate whether a bunched beam is to be used."""

    change_momentum: bool = True
    """Flag to indicate whether the wakefield can change the central momentum of the bunch."""

    factor: float = 1
    """
    Wake scaling factor.
    #TODO check if redundant based on scale_kick?
    """

    interpolate: bool = True
    """Flag to indicate whether to interpolate points in wake file."""

    scale_kick: float = 1
    """Factor by which to scale wake kicks."""

    t_column: str | None = None
    """t column in wake file"""

    z_column: str | None = None
    """z column in wake file"""

    wx_column: str | None = None
    """Wx column in wake file"""

    wy_column: str | None = None
    """Wy column in wake file"""

    wz_column: str | None = None
    """Wz column in wake file"""

    scale_field_ex: float = 0.0
    """x-component of the longitudinal direction vector."""

    scale_field_ey: float = 0.0
    """y-component of the longitudinal direction vector."""

    scale_field_ez: float = 1.0
    """z-component of the longitudinal direction vector."""

    scale_field_hx: float = 1.0
    """x-component of the horizontal direction vector."""

    scale_field_hy: float = 0.0
    """y-component of the horizontal direction vector."""

    scale_field_hz: float = 0.0
    """z-component of the horizontal direction vector."""

    equal_grid: float = 0.66
    """If 1.0 an equidistant grid is set up, if 0.0 a grid with equal charge per grid cell is
    employed. Values between 1.0 and 0.0 result in intermediate binning based on
    a linear combination of the two methods."""

    interpolation_method: int = 2
    """Interpolation method for ASTRA: 0 = rectangular, 1 = triangular, 2 = Gaussian."""
    smooth: float = 0.25
    """Smoothing parameter for Gaussian interpolation."""

    subbins: int = 10
    """Sub binning parameter."""

class TwissMatchSimulationElement(IgnoreExtra):
    beta_x: float
    """Horizontal beta"""

    beta_y: float
    """Vertical beta"""

    alpha_x: float
    """Horizontal alpha"""

    alpha_y: float
    """Vertical alpha"""

    eta_x: float = 0.0
    """Horizontal dispersion"""

    eta_y: float = 0.0
    """Vertical dispersion"""

    eta_xp: float = 0.0
    """Horizontal dispersion derivative"""

    eta_yp: float = 0.0
    """Vertical dispersion derivative"""

    from_beam: bool = True
    """If `True`, compute transformation from tracked beam properties instead of Twiss parameters"""


from nala.models.simulation import ApertureElement
from .base import BaseElementTranslator

class ApertureTranslator(BaseElementTranslator):
    aperture: ApertureElement

    def _write_ASTRA_Common(self, dic: dict) -> dict:
        """
        Creates the part of the ASTRA element dictionary common to all apertures in ASTRA

        Parameters
        ----------
        dic: dict
            Dictionary containing the parameters for the aperture

        Returns
        -------
        dict
            ASTRA dictionary with parameters and values
        """
        if self.aperture.negative_extent is not None:
            dic["Ap_Z1"] = {"value": self.aperture.negative_extent, "default": 0}
            dic["a_pos"] = {"value": self.physical.start.z}
        else:
            dic["Ap_Z1"] = {"value": self.physical.start.z + self.dz, "default": 0}
        if self.aperture.positive_extent is not None:
            dic["Ap_Z2"] = {"value": self.aperture.positive_extent, "default": 0}
            dic["a_pos"] = {"value": self.physical.start.z}
        else:
            end = (
                self.physical.end.z + self.dz
                if self.physical.end.z >= (self.physical.start.z + 1e-3)
                else self.physical.start.z + self.dz + 1e-3
            )
            dic["Ap_Z2"] = {"value": end, "default": 0}
        dic["A_xrot"] = {
            "value": self.x_rot + self.dx_rot,
            "default": 0,
            "type": "not_zero",
        }
        dic["A_yrot"] = {
            "value": self.y_rot + self.dy_rot,
            "default": 0,
            "type": "not_zero",
        }
        dic["A_zrot"] = {
            "value": self.z_rot + self.dz_rot,
            "default": 0,
            "type": "not_zero",
        }
        return dic

    def _write_ASTRA_Circular(self) -> dict:
        """
        Creates the part of the ASTRA element dictionary relevant to circular apertures in ASTRA

        Parameters
        ----------
        dic: dict
            Dictionary containing the parameters for the aperture

        Returns
        -------
        dict
            ASTRA dictionary with parameters and values
        """
        dic = dict()
        dic["File_Aperture"] = {"value": "RAD"}
        if self.aperture.radius is not None:
            radius = self.aperture.radius
        elif self.aperture.horizontal_size > 0 and self.aperture.vertical_size > 0:
            radius = min([self.aperture.horizontal_size, self.aperture.vertical_size])
        elif self.aperture.horizontal_size > 0:
            radius = self.aperture.horizontal_size
        elif self.aperture.vertical_size > 0:
            radius = self.aperture.vertical_size
        else:
            radius = 1
        dic["Ap_R"] = {"value": 1e3 * radius}
        return self._write_ASTRA_Common(dic)

    def _write_ASTRA_Planar(self, plane, width) -> dict:
        """
        Creates the part of the ASTRA element dictionary common to all apertures in ASTRA

        Parameters
        ----------
        dic: dict
            Dictionary containing the parameters for the aperture

        Returns
        -------
        dict
            ASTRA dictionary with parameters and values
        """
        dic = dict()
        dic["File_Aperture"] = {"value": plane}
        dic["Ap_R"] = {"value": width}
        return self._write_ASTRA_Common(dic)

    def to_astra(self, n: int = 0, **kwargs: dict) -> str:
        """
        Writes the aperture element string for ASTRA

        Parameters
        ----------
        n: int
            Element index number
        **kwargs: dict
            Keyword args

        Returns
        -------
        str
            String representation of the element for ASTRA

        Raises:
        -------
        ValueError
            If `shape` is not in the list of allowed values.
        """
        self.start_write()
        self.aperture.number_of_elements = 0
        if self.aperture.shape in ["elliptical", "circular"]:
            self.aperture.number_of_elements += 1
            dic = self._write_ASTRA_Circular()
            return self._write_ASTRA_dictionary(dic, n)
        elif self.aperture.shape in ["planar", "rectangular"]:
            text = ""
            if self.aperture.horizontal_size is not None and self.aperture.horizontal_size > 0:
                dic = self._write_ASTRA_Planar("Col_X", 1e3 * self.aperture.horizontal_size)
                text += self._write_ASTRA_dictionary(dic, n)
                self.aperture.number_of_elements += 1
            if self.aperture.vertical_size is not None and self.aperture.vertical_size > 0:
                dic = self._write_ASTRA_Planar("Col_Y", 1e3 * self.aperture.vertical_size)
                if self.aperture.number_of_elements > 0:
                    self.aperture.number_of_elements += 1
                    n = n + 1
                    text += "\n"
                text += self._write_ASTRA_dictionary(dic, n)
            return text
        elif self.aperture.shape == "scraper":
            text = ""
            if self.aperture.horizontal_size is not None and self.aperture.horizontal_size > 0:
                dic = self._write_ASTRA_Planar("Scr_X", 1e3 * self.aperture.horizontal_size)
                text += self._write_ASTRA_dictionary(dic, n)
                self.aperture.number_of_elements += 1
            if self.aperture.vertical_size is not None and self.aperture.vertical_size > 0:
                dic = self._write_ASTRA_Planar("Scr_Y", 1e3 * self.aperture.vertical_size)
                if self.aperture.number_of_elements > 0:
                    self.aperture.number_of_elements += 1
                    n = n + 1
                    text += "\n"
                text += self._write_ASTRA_dictionary(dic, n)
            return text
        else:
            raise ValueError(
                "shape must be in ['elliptical', 'planar', 'circular', 'rectangular', 'scraper']"
            )



from pydantic import BaseModel, ConfigDict, Field
from typing import Dict, List, Any
import numpy as np
from ...utils.classes import getGrids

section_header_text_ASTRA = {
    "&APERTURE": "LApert",
    "&CAVITY": "LEField",
    "&SOLENOID": "LBField",
    "&QUADRUPOLE": "LQuad",
    "&DIPOLE": "LDipole",
    "&WAKE": "LWAKE",
}

class astra_header(BaseModel):
    """
    Generic class for generating ASTRA namelists
    """

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
        validate_assignment=True,
        populate_by_name=True,
    )

    objectname: str = Field(alias="name")
    """Name of the object, used as a unique identifier in the simulation."""

    objecttype: str = Field(alias="type")
    """Type of the object, which determines its behavior and properties in the simulation."""

    header: str
    """Name of astra header"""

    exclude: List[str] = [
        "objectname",
        "objecttype",
        "astradict",
        "header",
        "exclude",
        "offset",
        "starting_offset",
        "starting_rotation",
        "global_parameters",
        "particle_definition",
        "output_particle_definition",
    ]

    astradict: Dict = {}

    def write_ASTRA(self) -> str:
        """
        Write the text for the ASTRA namelist based on its attributes.

        Returns
        -------
        str
            ASTRA-compatible string representing the namelist
        """
        output = f"&{self.header}\n"
        for key, val in self.model_dump().items():
            if key not in self.exclude and val is not None:
                if key in self.astradict:
                    output += f"{self.astradict[key]} = {val},\n"
                else:
                    output += f"{key} = {val},\n"
        output = output[:-2] + "\n"
        output += "/\n"
        return output


class astra_newrun(astra_header):
    """
    Class for generating the &NEWRUN namelist for ASTRA. See `ASTRA manual`_ for more details.
    """

    header: str = "NEWRUN"

    sample_interval: int = 1
    """Downsampling factor (as 2**(3 * sample_interval))"""

    run: int = 1
    """Run number"""

    head: str = "trial"
    """Run name"""

    lprompt: bool = False
    """If true a pause statement is included at the end
    of the run to avoid vanishing of the window in case of an error."""

    input_particle_definition: str
    """Name of input particle definition"""

    high_res: bool = True
    """If true, particle distributions are saved with increased accuracy."""

    auto_phase: bool = True
    """Phase RF cavities automatically"""

    bunch_charge: float | None = None
    """Bunch charge"""

    toffset: float = 0.0
    """Time offset of reference particle"""

    offset: list | np.ndarray = [0, 0, 0]
    """Beam offset from nominal axis [x,y,z]"""

    track_all: bool = True
    """If false, only the reference particle will be tracked"""

    phase_scan: bool = False
    """If true, the RF phases of the cavities will be scanned between 0 and 360 degree.
    Results are saved in the PScan file. The tracking between cavities will be done
    with the user-defined phases."""

    check_ref_part: bool = False
    """If true, the run will be interrupted if the reference particle is lost during the on-
    and off-axis reference particle tracking."""

    h_max: float = 0.07
    """Maximum time step for the Runge-Kutta integration."""

    h_min: float = 0.07
    """Minimum time step for the Runge-Kutta integration."""

    objectname: str = "newrun"
    """Name of object"""

    objecttype: str = "astra_newrun"
    """Type of object"""

    exclude: List = [
        "objectname",
        "objecttype",
        "astradict",
        "header",
        "exclude",
        "offset",
        "starting_offset",
        "starting_rotation",
        "global_parameters",
        "particle_definition",
        "output_particle_definition",
    ]

    def model_post_init(self, context: Any, /) -> None:
        self.astradict = {
            "input_particle_definition": "Distribution",
            "sample_interval": "n_red",
            "toffset": "Toff",
        }

    def write_ASTRA(self) -> str:
        if not self.input_particle_definition:
            raise ValueError("input_particle_definition must be defined for astra_newrun")
        return super().write_ASTRA()

class astra_output(astra_header):
    """
    Class for generating the &OUTPUT namelist for ASTRA. See `ASTRA manual`_ for more details.
    """

    header: str = "OUTPUT"

    lmagnetized: bool = False
    """If true, solenoid fields are neglected in the calculation of the beam emittance."""

    refs: bool = True
    """If true, output files according to Table 3 and Table 4 are generated. See `ASTRA manual`_"""

    emits: bool = True
    """If true, output files according to Table 3 and Table 4 are generated. See `ASTRA manual`_"""

    phases: bool = True
    """If true, output files according to Table 3 and Table 4 are generated. See `ASTRA manual`_"""

    high_res: bool = True
    """If true, particle distributions are saved with increased accuracy."""

    tracks: bool = True
    """If true, output files according to Table 3 and Table 4 are generated. See `ASTRA manual`_"""

    screens: List = []
    """List of :class:`~nala.models.element.Diagnostic` objects"""

    Lsub_cor: bool = True

    objectname: str = "output"
    """Name of object"""

    objecttype: str = "astra_output"
    """Type of object"""

    offset: list | np.ndarray = [0, 0, 0]
    """Beam offset from nominal axis [x,y,z]"""

    # starting_offset: float = None

    zstart: float = None

    zstop: float = None

    zemit: int = None

    # section: SectionLatticeTranslator

    def model_post_init(self, context: Any, /) -> None:
        self.astradict = {
            "input_particle_definition": "Distribution",
            "sample_interval": "n_red",
            "toffset": "Toff",
        }
        self.exclude.extend(["screens", "section", "end_element", "start_element"])#, "starting_offset"])
        # self.zstart = list(self.section.elements.elements.values())[0].physical.start.z
        # self.zstop = list(self.section.elements.elements.values())[-1].physical.end.z
        # self.zemit = int((self.zstop - self.zstart) / 0.01)
        # self.screens = [e for e in self.section.elements.elements.values() if e.hardware_class == "Diagnostic"]

    def write_ASTRA(self) -> str:
        """
        Write the text for the ASTRA namelist based on its :attr:`~framework_dict`.

        Parameters
        ----------
        n: int
            Index of the ASTRA element

        Returns
        -------
        str
            ASTRA-compatible string representing the namelist
        """
        output = f"&{self.header}\n"
        for key, val in self.model_dump().items():
            if key not in self.exclude and val is not None:
                if key in self.astradict:
                    output += f"{self.astradict[key]} = {val},\n"
                else:
                    output += f"{key} = {val},\n"
        for i, element in enumerate(self.screens, 1):
            output += f"Screen({i}) = {element.physical.middle.z},\n"
        output = output[:-2] + "\n"
        output += "/\n"
        return output

class astra_charge(astra_header):
    """
    Class for generating the &CHARGE namelist for ASTRA. See `ASTRA manual`_ for more details.
    """

    header: str = "CHARGE"

    npart: int = 2 ** (3 * 5)
    """Number of particles"""

    sample_interval: int = 1
    """Downsampling interval calculated as 2 ** (3 * sample_interval)"""

    space_charge_mode: str = "False"
    """Space charge mode"""

    space_charge_2D: bool = True
    """Enable 2D space charge calculations"""

    space_charge_3D: bool = False
    """Enable 3D space charge calculations"""

    cathode: bool = False
    """Flag to indicate whether the bunch was emitted from a cathode."""

    min_grid: float = 3.424657e-13
    """Minimum grid length during emission."""

    max_scale: float = 0.1
    """If one of the space charge scaling factors exceeds the limit 1± max_scale a new
    space charge calculation is initiated."""

    cell_var: float = 2
    """Variation of the cell height in radial direction."""

    nrad: int | None = None
    """Number of grid cells in radial direction up to the bunch radius."""

    nlong_in: int | None = None
    """Maximum number of grid cells in longitudinal direction within the bunch length."""

    smooth_x: int = 2
    """Smoothing parameter for x-direction. Only for 3D FFT algorithm."""

    smooth_y: int = 2
    """Smoothing parameter for y-direction. Only for 3D FFT algorithm."""

    smooth_z: int = 2
    """Smoothing parameter for z-direction. Only for 3D FFT algorithm."""

    grids: getGrids | None = None
    """Space charge grids"""

    objectname: str = "charge"
    """Name of object"""

    objecttype: str = "astra_charge"
    """Type of object"""

    def model_post_init(self, context: Any, /) -> None:
        self.grids = getGrids()
        self.astradict = {
            "cathode": "Lmirror",
            "space_charge_2D": "LSPCH",
            "space_charge_3D": "LSPCH3D",
        }
        self.exclude.extend(["grids", "npart", "sample_interval", "space_charge_mode", "mirror_charge"])

    def write_ASTRA(self) -> str:
        """
        Write the text for the ASTRA namelist based on its :attr:`~framework_dict`.

        Parameters
        ----------
        n: int
            Index of the ASTRA element

        Returns
        -------
        str
            ASTRA-compatible string representing the namelist
        """
        self.space_charge()
        output = f"&{self.header}\n"
        for key, val in self.model_dump().items():
            if key not in self.exclude and val is not None:
                if key in self.astradict:
                    output += f"{self.astradict[key]} = {val},\n"
                else:
                    output += f"{key} = {val},\n"
        if self.space_charge_2D:
            output += f"nrad = {self.grid_size},\n"
            output += f"nlong_in = {self.grid_size},\n"
        elif self.space_charge_3D:
            output += f"nxf = {self.grid_size},\n"
            output += f"nyf = {self.grid_size},\n"
            output += f"nzf = {self.grid_size},\n"
        elif self.space_charge():
            output += f"nxf = {self.grid_size},\n"
            output += f"nyf = {self.grid_size},\n"
            output += f"nzf = {self.grid_size},\n"
        output = output[:-2] + "\n"
        output += "/\n"
        return output

    def space_charge(self) -> bool:
        """
        Flag to indicate whether space charge is enabled.

        Returns
        -------
        bool
            True if enabled
        """
        if not (
            self.space_charge_mode == "False"
            or self.space_charge_mode is False
            or self.space_charge_mode is None
            or self.space_charge_mode == "None"
        ):
            self.space_charge_3D = True
            if isinstance(self.space_charge_mode, str):
                if "2d" in self.space_charge_mode.lower():
                    self.space_charge_2D = True
                    self.space_charge_3D = False
            return True
        self.space_charge_2D = False
        self.space_charge_3D = False
        return False


    @property
    def grid_size(self) -> int:
        """
        Get the number of space charge bins, see
        :func:`~SimulationFramework.Framework_objects.getGrids.getGridSizes`.

        Returns
        -------
        int
            The number of space charge bins based on the number of particles
        """
        # print('asking for grid sizes n = ', self.npart, ' is ', self.grids.getGridSizes(self.npart))
        return self.grids.getGridSizes(self.npart / self.sample_interval)

class astra_errors(astra_header):
    """
    Class for generating the &ERROR namelist for ASTRA. See `ASTRA manual`_ for more details.
    """

    header: str = "ERROR"

    global_errors: bool = True
    """If false, no errors will be generated."""

    log_error: bool = True
    """If true an additional log file will be generated which contains the actual
    element and bunch setting"""

    generate_output: bool = True
    """If true an output file will be generated"""

    suppress_output: bool = False
    """If true any generation of output other than the error file is suppressed."""

    objectname: str = "astra_error"
    """Name of object"""

    objecttype: str = "global_error"
    """Type of object"""

    def model_post_init(self, context: Any, /) -> None:
        self.exclude.extend(["element"])
        self.astradict = {
            "global_errors": "errors",
            "generate_output": "lerror",
        }


from pydantic import computed_field
import numpy as np
from .base import BaseElementTranslator
from nala.models.RF import RFCavityElement
from nala.models.simulation import RFCavitySimulationElement
from nala.translator.utils.fields import field

from ..converters import (
    elements_Elegant,
    elements_Opal,
)

class RFCavityTranslator(BaseElementTranslator):
    """
    Translator class for converting a :class:`~nala.models.element.RFCavity` element instance into a string or
    object that can be understood by various simulation codes.
    """

    cavity: RFCavityElement
    """Cavity element."""

    simulation: RFCavitySimulationElement
    """Cavity simulation element"""

    wakefile: str = None
    """Name of wakefile associated with the cavity."""

    trwakefile: str = None
    """Name of transverse wakefile associated with the cavity."""

    zwakefile: str = None
    """Name of longitudinal wakefile associated with the cavity."""

    @computed_field
    @property
    def structure_type(self) -> str:
        return self.cavity.structure_Type

    @computed_field
    @property
    def tcolumn(self) -> str | None:
        return f'"{self.simulation.t_column}"' if self.simulation.t_column else None

    @computed_field
    @property
    def zcolumn(self) -> str | None:
        return f'"{self.simulation.z_column}"' if self.simulation.z_column else None

    @computed_field
    @property
    def wxcolumn(self) -> str | None:
        return f'"{self.simulation.wx_column}"' if self.simulation.wx_column else None

    @computed_field
    @property
    def wycolumn(self) -> str | None:
        return f'"{self.simulation.wy_column}"' if self.simulation.wy_column else None

    @computed_field
    @property
    def wzcolumn(self) -> str | None:
        return f'"{self.simulation.wz_column}"' if self.simulation.wz_column else None


    def set_wakefield_column_names(self, wakefield_file_name: str) -> None:
        """
        Set the column names for the wakefield file, based on `wakefield_definition.

        Parameters
        ----------
        wakefield_file_name: str
            Name of the wakefield file
        """
        if all([x is not None for x in [self.wxcolumn, self.wycolumn, self.wzcolumn]]):
            self.wakefile = '"' + wakefield_file_name + '"'
            return
        elif self.wzcolumn is not None and all([x is None for x in [self.wxcolumn, self.wycolumn]]):
            self.zwakefile = '"' + wakefield_file_name + '"'
            return
        elif self.wzcolumn is None and all([x is not None for x in [self.wxcolumn, self.wycolumn]]):
            self.trwakefile = '"' + wakefield_file_name + '"'
            return

    def to_elegant(self) -> str:
        """
        Writes the cavity element string for ELEGANT.

        Returns
        -------
        str
            String representation of the element for ELEGANT
        """
        self.start_write()
        wholestring = ""
        etype = self._convertType_Elegant(self.hardware_type)
        if (
                self.simulation.wakefield_definition is None or
                self.simulation.wakefield_definition == ""
        ):
            etype = "rfca"
            # if self.simulation.field_definition is not None:
                # etype = "rftmez0"
                # if ".sdds" not in self.simulation.field_definition:
                #     field_file_name = self.generate_field_file_name(
                #     self.simulation.field_definition, code="elegant"
                # )
        else:
            wakefield_file_name = self.generate_field_file_name(
                self.simulation.wakefield_definition, code="elegant"
            )
            self.set_wakefield_column_names(wakefield_file_name)
        string = self.name + ": " + etype
        for key, value in self.full_dump().items():
            if (
                    not key == "name"
                    and not key == "type"
                    and not key == "commandtype"
                    and self._convertKeyword_Elegant(key, updated_type=self.hardware_type) in elements_Elegant[etype]
            ):
                if value is not None:
                    key = self._convertKeyword_Elegant(key, updated_type=self.hardware_type).lower()
                    # rftmez0 uses frequency instead of freq
                    if etype == "rftmez0" and key == "freq":
                        key = "frequency"

                    if self.hardware_type in ["RFCavity", "RFDeflectingCavity"]:
                        if key == "phase":
                            if etype == "rftmez0":
                                # If using rftmez0 or similar
                                value = (value / 360.0) * (2 * 3.14159)
                            else:
                                # In ELEGANT all phases are +90degrees!!
                                value = 90 - value

                    # In ELEGANT the voltages need to be compensated
                    if key == "volt":
                        if self.structure_type == "TravellingWave":
                            value = abs(
                                (self.get_cells() + 3.8)
                                * self.cavity.cell_length
                                * (1 / np.sqrt(2))
                                * value
                            )
                        else:
                            value = value
                    # If using rftmez0 or similar
                    if key == "ez_peak":
                        value = abs(1e-3 / (np.sqrt(2)) * value)

                    if key == "wakefile":
                        value = value

                    # In CAVITY NKICK = n_cells
                    if key == "n_kicks" and self.get_cells() > 1:
                        value = 3 * self.get_cells()

                    if key == "n_bins" and value > 0:
                        print(
                            "WARNING: Cavity n_bins is not zero - check log file to ensure correct behaviour!"
                        )
                    value = 1 if value is True else value
                    value = 0 if value is False else value
                    # print("elegant cavity", key, value)
                    tmpstring = ", " + key + " = " + str(value)
                # if len(string + tmpstring) > 156:
                #     wholestring += string + ",&\n"
                #     print(wholestring)
                #     string = ""
                #     string += tmpstring[2::]
                # else:
                    string += tmpstring
        wholestring += string + ";\n"
        return wholestring

    def to_ocelot(self) -> object:
        """
        Creates the cavity element for Ocelot.

        Returns
        -------
        tuple
            Ocelot Cavity object
        """
        from ..conversion_rules.codes import ocelot_conversion

        type_conversion_rules_Ocelot = ocelot_conversion.ocelot_conversion_rules
        self.start_write()
        self.generate_field_file_name(
            self.simulation.wakefield_definition, code="astra"
        )
        obj = type_conversion_rules_Ocelot[self.hardware_type](eid=self.name)
        for key, value in self.full_dump().items():
            if (
                    not key == "name"
                    and not key == "type"
                    and not key == "commandtype"
                    and self._convertKeyword_Ocelot(key) in obj.__class__().element.__dict__
            ):
                if value:
                    key = self._convertKeyword_Ocelot(key).lower()
                    if self.hardware_type in ["RFCavity", "RFDeflectingCavity"]:
                        if key == "v":
                            if self.structure_type == "TravellingWave":
                                value = (
                                        value
                                        * 1e-9
                                        * abs(
                                    (self.get_cells() + 3.8)
                                    * self.cavity.cell_length
                                    * (1 / np.sqrt(2))
                                )
                                )
                            else:
                                value = value * 1e-9
                    setattr(obj, key, value)
        return obj

    def to_cheetah(self) -> object:
        """
        Creates the cavity element for Cheetah.

        Returns
        -------
        tuple
            Cheetah Cavity object
        """
        from ..conversion_rules.codes import cheetah_conversion
        from torch import tensor, float64

        type_conversion_rules_Cheetah = cheetah_conversion.cheetah_conversion_rules
        self.start_write()
        obj = type_conversion_rules_Cheetah[self.hardware_type](
            name=self.name,
            length=tensor(self.physical.length, dtype=float64),
            sanitize_name=True
        )
        buffers = obj.__class__(length=tensor(self.physical.length, dtype=float64))._buffers
        for key, value in self.full_dump().items():
            if (key not in ["name", "type", "commandtype"]) and (
                    self._convertKeyword_Cheetah(key) in buffers
            ):
                key = self._convertKeyword_Cheetah(key)
                value = (
                    getattr(self, key)
                    if hasattr(self, key) and getattr(self, key) is not None
                    else value
                )
                if key == "voltage":
                    if self.structure_type == "TravellingWave":
                        value = (
                                value
                                * abs(
                            (self.get_cells() + 5.5)
                            * self.cavity.cell_length
                            * (1 / np.sqrt(2))
                        )
                        )
                    else:
                        value = value
                if isinstance(value, float):
                    dt = float64
                    setattr(obj, self._convertKeyword_Cheetah(key), tensor(value, dtype=dt))
                elif isinstance(value, int):
                    from torch import int64
                    dt = int64
                    setattr(obj, self._convertKeyword_Cheetah(key), tensor(value, dtype=dt))
        return obj

    def to_astra(self, n: int = 0, **kwargs: dict) -> str:
        """
        Writes the cavity element string for ASTRA.

        Parameters
        ----------
        n: int
            Element index number
        **kwargs: dict
            Keyword args

        Returns
        -------
        str
            String representation of the element for ASTRA
        """
        self.start_write()
        field_ref_pos = self.get_field_reference_position()
        auto_phase = kwargs["auto_phase"] if "auto_phase" in kwargs else True
        crest = self.cavity.crest if not auto_phase else 0
        field_file_name = self.generate_field_file_name(
            self.simulation.field_definition, code="astra"
        )
        efield_def = [
            "FILE_EFieLD",
            {"value": "'" + field_file_name + "'", "default": ""},
        ]
        return self._write_ASTRA_dictionary(
            dict(
                [
                    ["C_pos", {"value": field_ref_pos[2] + self.dz, "default": 0}],
                    efield_def,
                    ["C_numb", {"value": self.get_cells()}],
                    ["Nue", {"value": float(self.cavity.frequency) / 1e9, "default": 2998.5}],
                    [
                        "MaxE",
                        {"value": float(self.simulation.field_amplitude) / 1e6, "default": 0},
                    ],
                    ["Phi", {"value": crest - self.cavity.phase, "default": 0.0}],
                    ["C_smooth", {"value": self.simulation.smooth, "default": None}],
                    [
                        "C_xoff",
                        {
                            "value": field_ref_pos[0] + self.dx,
                            "default": None,
                            "type": "not_zero",
                        },
                    ],
                    [
                        "C_yoff",
                        {
                            "value": field_ref_pos[1] + self.dy,
                            "default": None,
                            "type": "not_zero",
                        },
                    ],
                    [
                        "C_xrot",
                        {
                            "value": self.x_rot + self.dx_rot,
                            "default": None,
                            "type": "not_zero",
                        },
                    ],
                    [
                        "C_yrot",
                        {
                            "value": self.y_rot + self.dy_rot,
                            "default": None,
                            "type": "not_zero",
                        },
                    ],
                    [
                        "C_zrot",
                        {
                            "value": self.z_rot + self.dz_rot,
                            "default": None,
                            "type": "not_zero",
                        },
                    ],
                ]
            ),
            n,
        )

    def to_xsuite(self, beam_length: int) -> tuple:
        """
        Generates an Xsuite object based on the element's properties and type.

        Parameters
        ----------
        beam_length: int
            Number of macroparticles in the beam

        Returns
        -------
        tuple
            (objectname, Xsuite object, properties[dict])
        """
        from ..conversion_rules.codes import xsuite_conversion

        type_conversion_rules_Xsuite = xsuite_conversion.xsuite_conversion_rules
        self.start_write()
        obj = type_conversion_rules_Xsuite[self.hardware_type]
        properties = {}
        for key, value in self.full_dump().items():
            if (key not in ["name", "type", "commandtype"]) and (
                    self._convertKeyword_Xsuite(key) in key in list(obj.__dict__.keys())
            ):
                key = self._convertKeyword_Xsuite(key)
                if key == "phase":
                    value = 90 - value
                if key == "field_amplitude":
                    if self.structure_type == "TravellingWave":
                        value = (
                            value
                            * abs(
                                (self.get_cells() + 3.8)
                                * self.cavity.cell_length
                                * (1 / np.sqrt(2))
                            )
                        )
                    else:
                        value = value
                if key == "n_kicks" and self.get_cells() > 1:
                    value = 3 * self.get_cells()
                properties.update({key: value})
        return self.name, obj, properties

    def to_opal(self, sval: float, designenergy: float | None = None) -> str:
        """
        Generates a string representation of the object's properties in the OPAL format.

        Parameters
        ----------
        sval: float
            S-position of the element
        designenergy: float, optional
            Beam energy at element in MeV

        Returns
        -------
        str
            A formatted string representing the object's properties in OPAL format.
        """
        self.start_write()
        etype = self._convertType_Opal(self.hardware_type)
        if self.structure_type == "TravellingWave":
            etype = "travelingwave"
        wholestring = self.name.replace('-', '_') + ": " + etype
        field_file_name = self.generate_field_file_name(
            self.simulation.field_definition, code="opal"
        )
        if etype.lower() == "drift" or self.simulation.field_definition is None:
            return ""
        for key, value in self.full_dump().items():
            if (
                    not key == "name"
                    and not key == "type"
                    and not key == "commandtype"
                    and self._convertKeyword_Opal(key) in elements_Opal[etype]
            ):
                if value is not None:
                    key = self._convertKeyword_Opal(key)
                    if key == "lag":
                        value = -value * np.pi / 180
                    if key == "freq":
                        value = value / 1e6
                    if key == "volt":
                        value = value / 1e6
                    val = 1 if value is True else value
                    val = 0 if value is False else val
                    if val is not None:
                        tmpstring = ", " + key + " = " + str(val)
                        wholestring += tmpstring
        if isinstance(self.simulation.field_definition, field):
            wholestring += ", fmapfn = \"" + self.generate_field_file_name(
                self.simulation.field_definition, code="opal"
            ) + "\""
            if self.structure_type == "TravellingWave":
                mode = float(self.simulation.field_definition.mode_numerator) / float(
                    self.simulation.field_definition.mode_denominator)
                wholestring += f", mode = {mode}"
        wholestring += f", ELEMEDGE = {sval};\n"
        return wholestring

    def get_cells(self) -> int:
        """
        Get the number of cavity cells.

        Returns
        -------
        int or None
            The number of cavity cells, or None if not defined.
        """
        if (self.cavity.n_cells == 0 or self.cavity.n_cells is None) and self.cavity.cell_length > 0:
            cells = round((self.physical.length - self.cavity.cell_length) / self.cavity.cell_length)
            cells = int(cells - (cells % 3))
        elif self.cavity.n_cells:
            if self.cavity.cell_length == self.physical.length:
                cells = 1
            else:
                cells = int(self.cavity.n_cells - (self.cavity.n_cells % 3))
        else:
            cells = 0
        return cells

    def to_gpt(self, Brho: float=0.0, ccs: str = "wcs", *args, **kwargs) -> str:
        """
        Write a string representation of the cavity for GPT

        #TODO note that not all possible ways of writing a cavity in GPT are currently supported.

        Parameters
        ----------
        Brho: float
            Magnetic rigidity; not used
        ccs: str
            Name of co-ordinate system of the cavity

        Returns
        -------
        str
            String representation of the cavity for GPT.
        """
        self.start_write()
        field_ref_pos = self.get_field_reference_position()
        ccs_label, value_text = self.ccs.ccs_text(field_ref_pos, self.physical.rotation.model_dump())
        relpos, _ = self.ccs.relative_position(field_ref_pos, self.physical.global_rotation.model_dump())
        field_file_name = self.generate_field_file_name(
            self.simulation.field_definition, code="gpt"
        )
        self.generate_field_file_name(self.simulation.wakefield_definition, code="gpt")
        """
        map1D_TM("wcs","z",linacposition,"mockup2m.gdf","Z","Ez",ffacl,phil,w);
        wakefield("wcs","z",  6.78904 + 4.06667 / 2, 4.06667, 50, "Sz5um10mm.gdf", "z","","","Wz", "FieldFactorWz", 10 * 122 / 4.06667) ;
        """
        subname = str(relpos[2]).replace(".", "")
        output = ""
        if field_file_name is not None:
            output = (
                    "f"
                    + subname
                    + " = "
                    + str(self.cavity.frequency)
                    + ";\n"
                    + "w"
                    + subname
                    + " = 2*pi*f"
                    + subname
                    + ";\n"
                    + "phi"
                    + subname
                    + " = "
                    + str((self.cavity.crest + 90 - self.cavity.phase + 0) % 360.0)
                    + "/deg;\n"
            )
            if self.structure_type == "TravellingWave":
                output += (
                        "ffac"
                        + subname
                        + " = 1.007 * "
                        + str((9.0 / (2.0 * np.pi)) * self.simulation.field_amplitude)
                        + ";\n"
                )
            else:
                output += "ffac" + subname + " = " + str(self.simulation.field_amplitude) + ";\n"

            # if False and self.Structure_Type == 'TravellingWave' and hasattr(self, 'attenuation_constant') and hasattr(self, 'shunt_impedance') and hasattr(self, 'design_power') and hasattr(self, 'design_gamma'):
            #     '''
            #     trwlinac(ECS,ao,Rs,Po,P,Go,thetao,phi,w,L)
            #     '''
            #     relpos, relrot = ccs.relative_position(self.middle, self.global_rotation)
            #     power = float(self.field_amplitude) / 25e6 * float(self.design_power)
            #     output += 'trwlinac' + '( ' + ccs.name + ', "z", '+ str(relpos[2]+self.coupling_cell_length) + ', ' + str(self.attenuation_constant / self.length) + ', ' + str(float(self.shunt_impedance) / self.length)\
            #             + ', ' + str(float(self.design_power) / self.length) + ', ' + str(power / self.length) + ', ' + str(1000/0.511) + ', ' + str(self.crest)\
            #             + ', '+str(self.phase)+', w'+subname+', ' + str(self.length) + ');\n'
            # else:
            output += (
                    "map1D_TM"
                    + "(\""
                    + self.ccs.name
                    + "\", "
                    + ccs_label
                    + ", "
                    + value_text
                    + ', "'
                    + str(field_file_name)
                    + '", "z", "Ez", ffac'
                    + subname
                    + ", phi"
                    + subname
                    + ", w"
                    + subname
                    + ");\n"
            )
        else:
            output = ""
        return output




import os

from ..models.elementList import MachineModel
from Importers.CATAP_Loader import catap_files, read_CATAP_YAML

if __name__ == "__main__":
    CATAPmachine = MachineModel(layout_file=os.path.abspath("./layouts.yaml"))

    for f in catap_files:
        elem = read_CATAP_YAML(f)
        CATAPmachine.update({n: e for n, e in elem.items()})



"""
Simple units functionality for the openPMD beamphysics records.

For more advanced units, use a package like Pint:
    https://pint.readthedocs.io/


"""

# import scipy.constants

m_e = 510998.94999999995
m_p = 938272088.16
c_light = 299792458
e_charge = 1.602176634e-19
mu_0 = 1.25663706212e-06


import numpy as np


class pmd_unit:
    """

    Params
    ------

    unitSymbol: Native units name
    unitSI:     Conversion factor to the the correspontign SI unit
    unitDimension: SI Base Exponents

    Base unit dimensions are defined as:
       Base dimension  | exponents.       | SI unit
       ---------------- -----------------   -------
       length          : (1,0,0,0,0,0,0)     m
       mass            : (0,1,0,0,0,0,0)     kg
       time            : (0,0,1,0,0,0,0)     s
       current         : (0,0,0,1,0,0,0)     A
       temperture      : (0,0,0,0,1,0,0)     K
       mol             : (0,0,0,0,0,1,0)     mol
       luminous        : (0,0,0,0,0,0,1)     cd

    Example:
        pmd_unit('eV', 1.602176634e-19, (2, 1, -2, 0, 0, 0, 0))
        defines that an eV is 1.602176634e-19 of base units m^2 kg/s^2, which is a Joule (J)

    If unitSI=0 (default), init with a known symbol:
        pmd_unit('T')
    returns:
        pmd_unit('T', 1, (0, 1, -2, -1, 0, 0, 0))


    Simple equalities are provided:
        u1 == u2
    Returns True if the params are all the same.

    """

    def __init__(self, unitSymbol="", unitSI=0, unitDimension=(0, 0, 0, 0, 0, 0, 0)):

        # Allow to return an internally known unit
        if unitSI == 0:
            if unitSymbol in known_unit:
                # Copy internals
                u = known_unit[unitSymbol]
                unitSI = u.unitSI
                unitDimension = u.unitDimension
            else:
                raise ValueError(f"unknown unitSymbol: {unitSymbol}")

        self._unitSymbol = unitSymbol
        self._unitSI = unitSI
        if isinstance(unitDimension, str):
            self._unitDimension = DIMENSION[unitDimension]
        else:
            self._unitDimension = unitDimension

    @property
    def unitSymbol(self):
        return self._unitSymbol

    @property
    def unitSI(self):
        return self._unitSI

    @property
    def unitDimension(self):
        return self._unitDimension

    def __mul__(self, other):
        return multiply_units(self, other)

    def __truediv__(self, other):
        return divide_units(self, other)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.unitSymbol

    def __repr__(self):
        return f"pmd_unit('{self.unitSymbol}', {self.unitSI}, {self.unitDimension})"


def is_identity(u):
    """Checks if the unit is equivalent to 1"""
    return u.unitSI == 1 and u.unitDimension == (0, 0, 0, 0, 0, 0, 0)


def multiply_units(u1, u2):
    """
    Multiplies two pmd_unit symbols
    """

    if is_identity(u1):
        return u2
    if is_identity(u2):
        return u1

    s1 = u1.unitSymbol
    s2 = u2.unitSymbol
    if s1 == s2:
        symbol = f"{s1}^2"
    else:
        symbol = s1 + "*" + s2
    d1 = u1.unitDimension
    d2 = u2.unitDimension
    dim = tuple(sum(x) for x in zip(d1, d2))
    unitSI = u1.unitSI * u2.unitSI

    return pmd_unit(unitSymbol=symbol, unitSI=unitSI, unitDimension=dim)


def divide_units(u1, u2):
    """
    Divides two pmd_unit symbols : u1/u2
    """

    if is_identity(u2):
        return u1

    s1 = u1.unitSymbol
    s2 = u2.unitSymbol
    if s1 == s2:
        symbol = "1"
    else:
        symbol = s1 + "/" + s2
    d1 = u1.unitDimension
    d2 = u2.unitDimension
    dim = tuple(a - b for a, b in zip(d1, d2))
    unitSI = u1.unitSI / u2.unitSI

    return pmd_unit(unitSymbol=symbol, unitSI=unitSI, unitDimension=dim)


def sqrt_unit(u):
    """
    Returns the sqrt of a unit
    """
    u.unitDimension

    symbol = u.unitSymbol
    if symbol not in ["", "1"]:
        symbol = f"sqrt({symbol})"

    unitSI = np.sqrt(u.unitSI)
    dim = tuple(x / 2 for x in u.unitDimension)

    return pmd_unit(unitSymbol=symbol, unitSI=unitSI, unitDimension=dim)


DIMENSION = {
    "1": (0, 0, 0, 0, 0, 0, 0),
    # Base units
    "length": (1, 0, 0, 0, 0, 0, 0),
    "mass": (0, 1, 0, 0, 0, 0, 0),
    "time": (0, 0, 1, 0, 0, 0, 0),
    "current": (0, 0, 0, 1, 0, 0, 0),
    "temperture": (0, 0, 0, 0, 1, 0, 0),
    "mol": (0, 0, 0, 0, 0, 1, 0),
    "luminous": (0, 0, 0, 0, 0, 0, 1),
    #
    "charge": (0, 0, 1, 1, 0, 0, 0),
    "electric_field": (1, 1, -3, -1, 0, 0, 0),
    "electric_potential": (1, 2, -3, -1, 0, 0, 0),
    "magnetic_field": (0, 1, -2, -1, 0, 0, 0),
    "velocity": (1, 0, -1, 0, 0, 0, 0),
    "energy": (2, 1, -2, 0, 0, 0, 0),
    "momentum": (1, 1, -1, 0, 0, 0, 0),
}
# Inverse
DIMENSION_NAME = {v: k for k, v in DIMENSION.items()}


def dimension(name):
    if name in DIMENSION:
        return DIMENSION[name]
    else:
        return None


def dimension_name(dim_array):
    return DIMENSION_NAME[tuple(dim_array)]


SI_symbol = {
    "1": "1",
    "length": "m",
    "mass": "kg",
    "time": "s",
    "current": "A",
    "temperture": "K",
    "mol": "mol",
    "luminous": "cd",
    "charge": "C",
    "electric_field": "V/m",
    "electric_potential": "V",
    "velocity": "m/s",
    "energy": "J",
    "momentum": "kg*m/s",
    "magnetic_field": "T",
}
# Inverse
SI_name = {v: k for k, v in SI_symbol.items()}


known_unit = {
    "1": pmd_unit("", 1, "1"),
    "rad": pmd_unit("", 1, "1"),
    "m": pmd_unit("m", 1, "length"),
    "kg": pmd_unit("kg", 1, "mass"),
    "g": pmd_unit("g", 0.001, "mass"),
    "s": pmd_unit("s", 1, "time"),
    "A": pmd_unit("A", 1, "current"),
    "K": pmd_unit("K", 1, "temperture"),
    "mol": pmd_unit("mol", 1, "mol"),
    "cd": pmd_unit("cd", 1, "luminous"),
    "C": pmd_unit("C", 1, "charge"),
    "charge_num": pmd_unit("charge #", 1, "charge"),
    "V/m": pmd_unit("V/m", 1, "electric_field"),
    "V": pmd_unit("V", 1, "electric_potential"),
    "c_light": pmd_unit("vel/c", c_light, "velocity"),
    "c": pmd_unit("vel/c", c_light, "velocity"),
    "m/s": pmd_unit("m/s", 1, "velocity"),
    "eV": pmd_unit("eV", e_charge, "energy"),
    "J": pmd_unit("J", 1, "energy"),
    "eV/c": pmd_unit("eV/c", e_charge / c_light, "momentum"),
    "T": pmd_unit("T", 1, "magnetic_field"),
}


def unit(symbol):
    """
    Returns a pmd_unit from a known symbol.

    * is allowed between two known symbols:
    """
    if symbol in known_unit:
        return known_unit[symbol]

    if "*" in symbol:
        subunits = [known_unit[s] for s in symbol.split("*")]
        # Require these to be in known units
        assert len(subunits) == 2, "TODO: more complicated units"
        return multiply_units(subunits[0], subunits[1])

    raise ValueError(f"Unknown unit symbol: {symbol}")


# Dicts for prefixes
PREFIX_FACTOR = {
    "yocto-": 1e-24,
    "zepto-": 1e-21,
    "atto-": 1e-18,
    "femto-": 1e-15,
    "pico-": 1e-12,
    "nano-": 1e-9,
    "micro-": 1e-6,
    "milli-": 1e-3,
    "centi-": 1e-2,
    "deci-": 1e-1,
    "deca-": 1e1,
    "hecto-": 1e2,
    "kilo-": 1e3,
    "mega-": 1e6,
    "giga-": 1e9,
    "tera-": 1e12,
    "peta-": 1e15,
    "exa-": 1e18,
    "zetta-": 1e21,
    "yotta-": 1e24,
}
# Inverse
PREFIX = dict((v, k) for k, v in PREFIX_FACTOR.items())

SHORT_PREFIX_FACTOR = {
    "y": 1e-24,
    "z": 1e-21,
    "a": 1e-18,
    "f": 1e-15,
    "p": 1e-12,
    "n": 1e-9,
    "µ": 1e-6,
    "m": 1e-3,
    "c": 1e-2,
    "d": 1e-1,
    "": 1,
    "da": 1e1,
    "h": 1e2,
    "k": 1e3,
    "M": 1e6,
    "G": 1e9,
    "T": 1e12,
    "P": 1e15,
    "E": 1e18,
    "Z": 1e21,
    "Y": 1e24,
}
# Inverse
SHORT_PREFIX = dict((v, k) for k, v in SHORT_PREFIX_FACTOR.items())


# Nice scaling


def nice_scale_prefix(scale):
    """
    Returns a nice factor and a SI prefix string

    Example:
        scale = 2e-10

        f, u = nice_scale_prefix(scale)


    """

    if scale == 0:
        return 1, ""

    p10 = np.log10(abs(scale))

    if p10 < -2 or p10 > 2:
        f = 10 ** (p10 // 3 * 3)
    else:
        f = 1

    return f, SHORT_PREFIX[f]


def nice_array(a):
    """
    Returns a scaled array, the scaling, and a unit prefix

    Example:
        nice_array( np.array([2e-10, 3e-10]) )
    Returns:
        (array([200., 300.]), 1e-12, 'p')

    """

    if np.isscalar(a):
        x = a
    elif len(a) == 1:
        x = a[0]
    else:
        a = np.array(a)
        x = a.ptp()

    fac, prefix = nice_scale_prefix(x)

    return a / fac, fac, prefix


# -------------------------
# Units for ParticleGroup

PARTICLEGROUP_UNITS = {}
for k in ["status"]:
    PARTICLEGROUP_UNITS[k] = unit("1")
for k in ["t"]:
    PARTICLEGROUP_UNITS[k] = unit("s")
for k in [
    "energy",
    "kinetic_energy",
    "mass",
    "higher_order_energy_spread",
    "higher_order_energy",
]:
    PARTICLEGROUP_UNITS[k] = unit("eV")
for k in ["px", "py", "pz", "p", "pr"]:
    PARTICLEGROUP_UNITS[k] = unit("eV/c")
for k in ["x", "y", "z", "r", "Jx", "Jy"]:
    PARTICLEGROUP_UNITS[k] = unit("m")
for k in ["beta", "beta_x", "beta_y", "beta_z", "gamma", "theta", "ptheta"]:
    PARTICLEGROUP_UNITS[k] = unit("1")
for k in ["charge", "species_charge", "weight"]:
    PARTICLEGROUP_UNITS[k] = unit("C")
for k in ["average_current"]:
    PARTICLEGROUP_UNITS[k] = unit("A")
for k in ["norm_emit_x", "norm_emit_y"]:
    PARTICLEGROUP_UNITS[k] = unit("m")
for k in ["norm_emit_4d"]:
    PARTICLEGROUP_UNITS[k] = multiply_units(unit("m"), unit("m"))
for k in ["xp", "yp"]:
    PARTICLEGROUP_UNITS[k] = unit("rad")
for k in ["x_bar", "px_bar", "y_bar", "py_bar"]:
    PARTICLEGROUP_UNITS[k] = sqrt_unit(unit("m"))


def pg_units(key):
    """
    Returns a str representing the units of any attribute
    """
    for prefix in ["sigma_", "mean_", "min_", "max_", "ptp_", "delta_"]:
        if key.startswith(prefix):
            nkey = key[len(prefix) :]
            return PARTICLEGROUP_UNITS[nkey]

    if key.startswith("cov_"):
        subkeys = key.strip("cov_").split("__")
        unit0 = PARTICLEGROUP_UNITS[subkeys[0]]

        unit1 = PARTICLEGROUP_UNITS[subkeys[1]]

        return multiply_units(unit0, unit1)

    # Fields
    if key.startswith("electricField"):
        return unit("V/m")
    if key.startswith("magneticField"):
        return unit("T")

    return PARTICLEGROUP_UNITS[key]


# -------------------------
# h5 tools


def write_unit_h5(h5, u):
    """
    Writes an pmd_unit to an h5 handle
    """

    h5.attrs["unitSI"] = u.unitSI
    h5.attrs["unitDimension"] = u.unitDimension
    h5.attrs["unitSymbol"] = u.unitSymbol


def read_unit_h5(h5):
    """
    Reads unit data from an h5 handle and returns a pmd_unit object
    """
    a = h5.attrs

    unitSI = a["unitSI"]
    unitDimension = tuple(a["unitDimension"])
    if "unitSymbol" not in a:
        unitSymbol = "unknown"
    else:
        unitSymbol = a["unitSymbol"]

    return pmd_unit(unitSymbol=unitSymbol, unitSI=unitSI, unitDimension=unitDimension)


def read_dataset_and_unit_h5(h5, expected_unit=None, convert=True):
    """
    Reads a dataset that has openPMD unit attributes.

    expected_unit can be a pmd_unit object, or a known unit str. Examples: 'kg', 'J', 'eV'

    If expected_unit is given, will check that the units are compatible.

    If convert, the data will be returned with the expected_units.


    Returns a tuple:
        np.array, pmd_unit

    """

    # Read the unit that is there.
    u = read_unit_h5(h5)

    # Simple case
    if not expected_unit:
        return np.array(h5), u

    if isinstance(expected_unit, str):
        # Try to get unit
        expected_unit = unit(expected_unit)

    # Check dimensions
    du = divide_units(u, expected_unit)

    assert du.unitDimension == (0, 0, 0, 0, 0, 0, 0), "incompatible units"

    if convert:
        fac = du.unitSI
        return fac * np.array(h5), expected_unit
    else:
        return np.array(h5), u


def write_dataset_and_unit_h5(h5, name, data, unit=None):
    """
    Writes data and pmd_unit to h5[name]

    See: read_dataset_and_unit_h5
    """
    h5[name] = data

    if unit:
        write_unit_h5(h5[name], unit)



from typing import List, Dict

from nala.models.element import (
    Element,
    Magnet,
    Solenoid,
    Dipole,
    RFCavity,
    RFDeflectingCavity,
    Drift,
    Aperture,
    Diagnostic,
    Marker,
    Plasma,
    Laser,
    Wiggler,
    Combined_Corrector,
    Horizontal_Corrector,
    Vertical_Corrector,
    NonLinearLens,
    TwissMatch,
)

from .base import BaseElementTranslator
from .magnet import (
    MagnetTranslator,
    SolenoidTranslator,
    DipoleTranslator,
    WigglerTranslator,
    NonLinearLensTranslator,
)
from .cavity import RFCavityTranslator
from .drift import DriftTranslator
from .diagnostic import DiagnosticTranslator
from .aperture import ApertureTranslator
from .plasma import PlasmaTranslator
from .laser import LaserTranslator
from .twiss import TwissMatchTranslator


def translate_elements(
        elements: List[Element],
        master_lattice_location: str = None,
        directory: str = '.',
) -> Dict[str, BaseElementTranslator]:
    """
    Function for translating a list of elements into their respective Translator classes.

    Parameters
    ----------
    elements: List[Element]
        List of :class:`~nala.models.element.Element` objects.
    master_lattice_location: str
        Directory containing lattice/data files including field/wakefield files.
    directory:
        Directory to which files will be written.

    Returns
    -------
    Dict[str, BaseElementTranslator]
        Dictionary of :class:`~nala.translator.converters.base.BaseElementTranslator` objects, keyed
        by their original name.
    """
    elem_dict = {}
    for elem in elements:
        if isinstance(elem, Magnet):
            if isinstance(elem, Solenoid):
                translator = SolenoidTranslator
            elif isinstance(elem, Dipole) and not type(elem) in [
                Combined_Corrector,
                Horizontal_Corrector,
                Vertical_Corrector,
            ]:
                translator = DipoleTranslator
            elif isinstance(elem, Wiggler):
                translator = WigglerTranslator
            elif isinstance(elem, NonLinearLens):
                translator = NonLinearLensTranslator
            else:
                translator = MagnetTranslator
        elif type(elem) in [RFCavity, RFDeflectingCavity]:
            translator = RFCavityTranslator
        elif isinstance(elem, Drift):
            translator = DriftTranslator
        elif isinstance(elem, Diagnostic) or isinstance(elem, Marker):
            translator = DiagnosticTranslator
        elif isinstance(elem, Aperture):
            translator = ApertureTranslator
        elif isinstance(elem, Plasma):
            translator = PlasmaTranslator
        elif isinstance(elem, Laser):
            translator = LaserTranslator
        elif isinstance(elem, TwissMatch):
            translator = TwissMatchTranslator
        else:
            translator = BaseElementTranslator
        elem_dict.update({elem.name: translator.model_validate(elem.model_dump())})
        elem_dict[elem.name].master_lattice_location = master_lattice_location
        elem_dict[elem.name].directory = directory
    return elem_dict